"""
Method for SV characterization

Copyright (C) 2021 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""


import os
import logging
import random
from nanovar.nv_bam_parser import bam_parse
# from nanovar.nv_detect_algo import sv_detect
# from nanovar.nv_parser import entry_parser, align_info, breakpoint_parser
from nanovar.nv_cluster import sv_cluster
from nanovar.nv_nn import inference, svread_ovl
from nanovar.nv_cov_upper import ovl_upper
from nanovar.nv_vcf import create_vcf
from nanovar.nv_report import create_report
from nanovar.nv_dup_te_detect import dup_te_analyzer
from nanovar.nv_alt_seq import get_alt_seq
#from cytocad.change_detection import cad
#from cytocad.ideogram import tagore_wrapper


class VariantDetect:

    def __init__(self, wk_dir, bam, splitpct, minalign, filter_path, minlen, buff, model_path, total_gsize,
                 contig_len_dict, thres, read_path, read_name, ref_path, ref_name, map_cmd, mincov, homo_t, het_t, debug,
                 contig_omit, cnv, nv_cmd, sam):
        self.dir = wk_dir
        self.bam = bam
        self.sam = sam
        self.splitpct = splitpct
        self.minalign = minalign
        self.filter = filter_path
        self.contig_omit = contig_omit
        self.minlen = minlen
        self.buff = buff
        self.model = model_path
        self.gsize = total_gsize
        self.contig = contig_len_dict
        self.thres = thres
        self.rpath = read_path
        self.rname = read_name
        self.refpath = ref_path
        self.refname = ref_name
        self.mapcmd = map_cmd
        self.mincov = mincov
        self.homo_t = homo_t
        self.het_t = het_t
        self.debug = debug
        self.cnv = cnv
        self.basecov, self.maxovl, self.depth, self.maxovl3 = 0, 0, 0, 0
        self.total_out, self.total_subdata, self.out_nn, self.ins_out, self.out_rest, self.detect_out, self.beddata = [], [], [], [], [], [], []
        self.rlendict, self.parse_dict, self.index2te = {}, {}, {}
        # HTML SV table entry limit
        self.num_limit = 1000
        # HTML SV table SV ratio limit
        self.ratio_limit = 1
        self.maps = 0
        self.seed = 0
        self.seed2 = 1
        self.nv_cmd = nv_cmd

    def bam_parse_detect(self):
        random.seed(1)
        self.total_subdata, self.total_out, self.basecov, self.parse_dict, self.rlendict, self.maps, self.detect_out, self.seed, self.beddata \
            = bam_parse(self.sam, self.minlen, self.splitpct, self.minalign, self.dir, self.filter, self.contig_omit)
        writer(os.path.join(self.dir, 'subdata.tsv'), self.total_subdata, self.debug)
        writer(os.path.join(self.dir, 'detect.tsv'), self.detect_out, self.debug)
        writer(os.path.join(self.dir, 'parse1.tsv'), self.total_out, self.debug)

    def coverage_stats(self):
        # Obtaining upper cov limit and depth of coverage
        self.maxovl, self.depth, self.maxovl3 = ovl_upper(self.gsize, self.contig, self.basecov, self.beddata, self.dir)
        # CNV detection using cytocad
        if self.cnv:
            cnv_out, tag = cad(self.total_subdata, self.depth, self.maxovl3, self.rname, ref_build=self.cnv, cov_plots=True,
                               wk_dir=self.dir)
            tagore_wrapper(tag, self.rname, self.dir, self.cnv, 'png')
            cnv_out_path = os.path.join(self.dir, self.rname + ".CNV.bed")
            outwrite = open(cnv_out_path, 'w')
            _ = outwrite.write('\n'.join(cnv_out))
            outwrite.close()
            logging.info("CNV analysis completed.")
        # Report statistics of coverage
        logging.info("Genome size: %s bases" % str(self.gsize))
        logging.info("Mapped bases: %s bases" % str(self.basecov))
        logging.info("Depth of coverage: %sx" % str(self.depth))
        if self.depth < 4:
            logging.warning("Sequencing depth is less than 4x, output may not be comprehensive")
        logging.info("Read overlap upper limit: %s\n" % str(self.maxovl))

    def cluster_nn2(self):
        logging.info("Clustering SV breakends")
        # cluster_out, _ = sv_cluster(self.beddata, self.total_out, self.buff, self.maxovl, self.mincov,
        #                            self.contig, True, self.seed2)
        # Removed maxovl filter
        cluster_out, _ = sv_cluster(self.beddata, self.total_out, self.buff, 1000000, self.mincov,
                                    self.contig, True, self.seed2)
        logging.info("Neural network inference")
        if cluster_out:
            self.out_nn = inference(cluster_out, self.total_out, self.model)
        else:
            self.out_nn = []
        # Generate sv_overlap file
        #svread_ovl(self.dir, self.out_nn)

    # def cluster_extract(self):
    #     logging.info("Clustering SV breakends")
    #     cluster_parse = sv_cluster(self.total_subdata, self.total_out, self.buff, self.maxovl, self.mincov,
    #                                self.contig, False, self.seed2)
    #     logging.info("Filtering INS and INV SVs")
    #     total_qnames = []
    #     for line in cluster_parse:
    #         svtype = line.split('~')[1]
    #         qnames = []
    #         if svtype in ['bp_Nov_Ins', 'Nov_Ins']:  # Some INS are actually DUP
    #             for r in cluster_parse[line]:
    #                 qnames.append(r.rsplit('~', 1)[0])
    #                 self.ins_out.append(self.parse_dict[r])
    #             total_qnames.extend(qnames)
    #         elif svtype == 'Del':
    #             for r in cluster_parse[line]:
    #                 self.ins_out.append(self.parse_dict[r])
    #         else:
    #             for r in cluster_parse[line]:
    #                 qnames.append(r.rsplit('~', 1)[0])
    #             total_qnames.extend(qnames)
    #     qnames_dict = {x: 1 for x in set(total_qnames)}
    #     self.fasta_extract(qnames_dict)

    # def cluster_nn(self, add_out):
    #     logging.info("Re-clustering INS/INV SVs and merging")
    #     # Merge old ins with new from hsblastn
    #     sub_out = self.ins_out + add_out
    #     cluster_out_ins, _ = sv_cluster(self.total_subdata, sub_out, self.buff, self.maxovl, self.mincov, self.contig, True,
    #                                     self.seed2)
    #     new_cluster_out = self.out_rest + cluster_out_ins
    #     logging.info("Neural network inference")
    #     new_total_out = self.total_out + add_out
    #     if new_cluster_out:
    #         self.out_nn = inference(new_cluster_out, new_total_out, self.model)
    #     else:
    #         self.out_nn = []
    #     # Generate sv_overlap file
    #     svread_ovl(self.dir, self.out_nn)

    # def parse_detect_hsb(self):
    #     # Make gap dictionary
    #     gapdict = makegapdict(self.filter, self.contig_omit)
    #     temp1 = []
    #     chromocollect = []
    #     ovlt = 0.9
    #     sig_index = [0, 2, 4, 6, 8]
    #     data = open(self.bam, 'r').read().splitlines()
    #     data.append('null\tnull\tnull\tnull\tnull\tnull')
    #     nlines = len(data) - 1
    #     for i in range(nlines):
    #         if data[i].split('\t')[0] == data[i + 1].split('\t')[0]:  # Grouping alignments by read name
    #             temp1.append(align_info(data[i], self.rlendict))
    #             if data[i].split('\t')[1].strip() not in chromocollect:
    #                 chromocollect.append(data[i].split('\t')[1].strip())
    #         else:
    #             temp1.append(align_info(data[i], self.rlendict))
    #             if data[i].split('\t')[1].strip() not in chromocollect:
    #                 chromocollect.append(data[i].split('\t')[1].strip())
    #             self.seed += 1
    #             # Parse entries and correct overlap alignments
    #             subdata = entry_parser(temp1, chromocollect, ovlt)
    #             # SV detection
    #             out1, out2 = sv_detect(subdata, self.splitpct, self.minalign, gapdict)
    #             if out2 == '':
    #                 pass
    #             else:
    #                 # Parse breakpoints
    #                 final = breakpoint_parser(out2, self.minlen, sig_index, self.seed, 'hsb')
    #                 self.total_out.extend(final)
    #             temp1 = []
    #             chromocollect = []
    #     if not self.debug:  # Remove blast table if not debug mode
    #         os.remove(self.bam)
    
    def dup_te_detect(self, ref_dir, threads, mm, st, data_type):
        logging.info("Detecting DUP and TE")
        self.index2te, self.out_nn = dup_te_analyzer(self.dir, self.out_nn, self.total_out, self.thres, ref_dir, self.refpath, mm, threads, data_type, st, self.debug)
    
    def vcf_report(self):
        logging.info("Creating VCF")
        alt_seq = get_alt_seq(self.dir, self.out_nn, self.refpath)
        create_vcf(self.dir, self.thres, self.out_nn, self.refpath, self.rpath, self.rname, self.mapcmd, self.contig,
                   self.homo_t, self.het_t, self.minlen, self.depth, self.index2te, self.nv_cmd, alt_seq)
        logging.info("Creating HTML report")
        create_report(self.dir, self.contig, self.thres, self.rpath, self.refpath, self.rlendict, self.rname,
                      self.num_limit, self.ratio_limit)

    # Write intermediate data to file if debug mode
    def write2file(self, add_out):
        writer(os.path.join(self.dir, 'parse2.tsv'), add_out, self.debug)
        writer(os.path.join(self.dir, 'cluster.tsv'), self.out_nn, self.debug)

    # FASTA extractor
    def fasta_extract(self, qnames):
        fasta = os.path.join(self.dir, 'temp1.fa')
        out = open(os.path.join(self.dir, 'temp2.fa'), 'a')
        with open(fasta) as f:
            for line in f:
                try:
                    if qnames[line[1:].strip('\n')]:
                        line1 = line
                        line2 = next(f)
                        out.write(line1 + line2)
                except KeyError:
                    _ = next(f)
            out.close()
        # os.remove(fasta)


# Write to files
def writer(file_path, data_list, switch):
    if switch is True:
        w = open(file_path, 'w')
        w.write('\n'.join(data_list) + '\n')
        w.close()


# Create genome gap dictionary
def makegapdict(gap_path, contig_omit):
    gapdict = {}
    if gap_path is not None:
        rgapdata = open(gap_path, 'r').read().splitlines()
        for line in rgapdata:
            try:
                gapdict[line.split('\t')[0]].append([int(line.split('\t')[1]), int(line.split('\t')[2])])
            except KeyError:
                gapdict[line.split('\t')[0]] = [[int(line.split('\t')[1]), int(line.split('\t')[2])]]
    for contig in contig_omit:
        try:
            gapdict[contig].append(contig_omit[contig])
        except KeyError:
            gapdict[contig] = [contig_omit[contig]]
    if gapdict:  # Gap file is used or reference genome contigs contain invalid symbols
        logging.info('Gap dictionary successfully loaded')
        return gapdict
    else:
        logging.info('Gap dictionary not loaded.')
        return None



"""
Functions to create HTML report.

Copyright (C) 2021 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import math
import datetime
import numpy as np
import nanovar
import matplotlib
# import htmlark
import base64
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.ticker import ScalarFormatter
from distutils.dir_util import copy_tree
from nanovar import __version__
import base64
import mimetypes
from bs4 import BeautifulSoup
from urllib.parse import quote

# Create HTML report
def create_report(wk_dir, contig_len_dict, thres, read_path, ref_path, rlen_dict, read_name, num_limit, ratio_limit):
    timenow = datetime.datetime.now()
    fwd = os.path.join(os.getcwd(), wk_dir, 'fig')
    fwd_fig = './fig'
    threshold = thres
    len_cap = 100000
    num_header = len(contig_len_dict) + 28  # This has to adjusted according to number of headers in VCF
    # vcf_path = os.path.join(wk_dir, '%s.nanovar.total.vcf' % read_name)
    vcf_path_pass = os.path.join(wk_dir, '%s.nanovar.pass.vcf' % read_name)
    vcf_data = open(vcf_path_pass, 'r').read().splitlines()
    vcf = sorted(vcf_data[num_header:], key=lambda x: float(x.split('\t')[5]), reverse=True)
    # Creating variables
    scorelist, ratiolist, lcovlist = [], [], []
    svdict = {'DEL': 0, 'INV': 0, 'DUP': 0, 'INS': 0, 'BND': 0}
    dellen, inslen, invlen, duplen, bndlen = [], [], [], [], []
    delnolen, insnolen, invnolen, dupnolen, bndnolen = 0, 0, 0, 0, 0
    delcap, inscap, invcap, dupcap, bndcap = 0, 0, 0, 0, 0
    data = []
    totalsv, n = 0, 0
    for i in vcf:
        totalsv += 1
        scorelist.append(float(i.split('\t')[5]))
        lcov = int(i.split('\t')[9].split(':')[2].split(',')[1])
        lcovlist.append(lcov)
        ncov = int(i.split('\t')[9].split(':')[2].split(',')[0])
        ratio = round(lcov/(lcov + ncov), 3)
        ratiolist.append(ratio)
        if float(i.split('\t')[5]) >= threshold:
            if ratio <= ratio_limit:
                n += 1
                sv_type = i.split('\t')[2].split('.')[1]
                if sv_type != 'BND':
                # if i.split('\t')[4].startswith('<'):
                    # sv_type = i.split('\t')[4].strip('<>')
                    sv_len = i.split('\t')[7].split(';')[2].split('=')[1]
                    data.append([str(n), sv_type, i.split('\t')[0], i.split('\t')[1],
                                 '-', i.split('\t')[7].split(';')[1].split('=')[1],
                                 sv_len, i.split('\t')[5],
                                 str(lcov), str(ncov), str(ratio), i.split('\t')[9].split(':')[0],
                                 i.split('\t')[2]])
                    svdict[sv_type] += 1
                    if sv_type == 'DEL':
                        try:
                            if int(sv_len) * -1 <= len_cap:
                                dellen.append(svlencap(int(sv_len)) * -1)
                            else:
                                delcap += 1
                        except ValueError:
                            delnolen += 1
                    elif sv_type == 'INS':
                        try:
                            if int(sv_len) <= len_cap:
                                inslen.append(svlencap(int(sv_len)))
                            else:
                                inscap += 1
                        except ValueError:
                            insnolen += 1
                    elif sv_type == 'INV':
                        try:
                            if int(sv_len) <= len_cap:
                                invlen.append(svlencap(int(sv_len)))
                            else:
                                invcap += 1
                        except ValueError:
                            invnolen += 1
                    elif sv_type == 'DUP':
                        try:
                            if int(sv_len) <= len_cap:
                                duplen.append(svlencap(int(sv_len)))
                            else:
                                dupcap += 1
                        except ValueError:
                            dupnolen += 1
                    else:
                        raise Exception('Error: Unrecognized SV type %s' % sv_type)
                else:
                    sv_type = i.split('\t')[7].split(';')[5].split('=')[1]
                    chrom2 = i.split('\t')[4].split(':')[0].strip('N[]')
                    pos2 = i.split('\t')[4].split(':')[1].strip('N[]')
                    sv_len = i.split('\t')[7].split(';')[2].split('=')[1]
                    data.append([str(n), sv_type, i.split('\t')[0], i.split('\t')[1],
                                 chrom2, pos2, sv_len, i.split('\t')[5], str(lcov), str(ncov),
                                 str(ratio), i.split('\t')[9].split(':')[0], i.split('\t')[2]])
                    svdict['BND'] += 1
                    try:
                        if int(sv_len) == 0:
                            bndnolen += 1
                        elif int(sv_len) <= len_cap:
                            bndlen.append(svlencap(int(sv_len)))
                        else:
                            bndcap += 1
                    except ValueError:
                        bndnolen += 1
    totalsvlen = np.array([dellen, inslen, invlen, bndlen, duplen], dtype=object)
    tab = [[len(dellen), len(inslen), len(invlen), len(bndlen), len(duplen)], [delcap, inscap, invcap, bndcap, dupcap],
           [delnolen, insnolen, invnolen, bndnolen, dupnolen]]
    # Setting global figure parameters
    params = {'axes.labelsize': 14, 'axes.titlesize': 17, 'legend.fontsize': 10,
              'xtick.labelsize': 12, 'ytick.labelsize': 12, 'font.family': 'sans-serif'}
    matplotlib.rcParams.update(params)
    # Make plots
    scatter_plots(fwd, scorelist, ratiolist, lcovlist, threshold)
    sv_len_dict(fwd, totalsvlen, tab)
    sv_type_dist(svdict, fwd)
    read_len_dict(rlen_dict, fwd)
    # Resize data for table
    data2 = data[0:num_limit]
    # Get CSS and JS paths
    css = os.path.join(os.path.dirname(nanovar.__file__), 'css')
    js = os.path.join(os.path.dirname(nanovar.__file__), 'js')
    # Write HTML
    create_html(data2, fwd, wk_dir, vcf_path_pass, timenow, read_name, read_path, ref_path, threshold, n, totalsv, css, js)
    #css_to = os.path.join(os.getcwd(), wk_dir, 'css')
    #js_to = os.path.join(os.getcwd(), wk_dir, 'js')
    #null = copy_tree(css, css_to)
    #null = copy_tree(js, js_to)


# SV Length cap
def svlencap(x):
    # if x >= 10000:
    #     return 10000
    # else:
    #     return x
    return x


# Scatter plots generation
def scatter_plots(fwd, scorelist, ratiolist, lcovlist, threshold):
    # Scatter plot Score vs Read ratio
    fig = plt.figure(figsize=(8, 6))
    fig.patch.set_facecolor('#f6f7f9')
    ax = fig.add_subplot(111)
    ax.scatter(ratiolist, scorelist, c='#7d3f5d', alpha=0.1)
    ax.axhline(y=threshold, linewidth=1, color='firebrick')
    ax.annotate("Threshold=" + str(threshold), xy=(0.8, threshold+0.2))
    ax.set_facecolor('#ebebff')
    plt.ylabel('Confidence score')
    plt.xlabel('Breakend read ratio')
    plt.ylim(bottom=-0.3)
    plt.savefig(os.path.join(fwd, 'scatter1.png'), bbox_inches='tight', dpi=100, facecolor=fig.get_facecolor(), edgecolor='none')
    # Scatter plot Score vs Supporting reads
    fig = plt.figure(figsize=(8, 6))
    fig.patch.set_facecolor('#f6f7f9')
    ax = fig.add_subplot(111)
    if lcovlist:
        mean_cov = sum(lcovlist)/len(lcovlist)
        buffer = 10
        max_cov = max(mean_cov + buffer, 10)
        min_cov = max(mean_cov - buffer, 0)
    else:
        max_cov = 10
        min_cov = 0
    ax.scatter(lcovlist, scorelist, c='#3f5d7d', alpha=0.1)
    ax.axhline(y=threshold, linewidth=1, color='firebrick')
    ax.annotate("Threshold=" + str(threshold), xy=(max_cov - max_cov/5, threshold+0.2))
    ax.set_facecolor('#ebebff')
    plt.ylabel('Confidence score')
    plt.xlabel('Number of breakend-supporting reads')
    plt.ylim(bottom=-0.3)
    plt.xlim(left=min_cov, right=max_cov)
    plt.savefig(os.path.join(fwd, 'scatter2.png'), bbox_inches='tight', dpi=100, facecolor=fig.get_facecolor(), edgecolor='none')


# Boxplot SV length distribution
def sv_len_dict(fwd, totalsvlen, tab):
    fig = plt.figure(figsize=(8, 6))
    fig.patch.set_facecolor('#f6f7f9')
    ax = fig.add_subplot(111)
    label = ["Deletion", "Insertion", "Inversion", "Breakend", "TandemDup"]
    colors = ['#7d3f5d', '#3f5d7d', '#7d5f3f', '#403f7d', '#3f7c7d']
    medianprop = dict(linewidth=1.5, color='black')
    bp = ax.boxplot(totalsvlen, whis=[5, 95], medianprops=medianprop, showfliers=False, patch_artist=True)
    ax.tick_params(axis='x', which='both', bottom=False, labelbottom=False)
    for patch, color in zip(bp['boxes'], colors):
        patch.set_facecolor(color)
    ax.table(cellText=tab, rowLabels=['Size <= 100,000 bp', 'Size > 100,000 bp', 'Unknown size'],
             colLabels=label, loc='bottom', bbox=[0, -0.3, 1, 0.275])
    ax.set_facecolor('#ebebff')
    ax.xaxis.grid(False)
    ax.yaxis.grid(color='white', linewidth=1)
    if not all(len(x) == 0 for x in totalsvlen):  # Fix ValueError for log scale when no data
        ax.set_yscale('log')
    plt.yticks([100, 1000, 10000, 100000], ['100', '1,000', '10,000', '100,000'], rotation=45)
    plt.ylabel('SV length (bp)')
    plt.savefig(os.path.join(fwd, 'sv_lengths.png'), bbox_inches='tight', dpi=100, facecolor=fig.get_facecolor(),
                edgecolor='none')


# Donut SV types
def sv_type_dist(svdict, fwd):
    fig, ax = plt.subplots(figsize=(10, 5), subplot_kw=dict(aspect="equal"))
    fig.patch.set_facecolor('#f6f7f9')
    svnamedict = {"DEL": " Deletions", "INS": " Insertions",
                  "INV": " Inversions", "DUP": " TandemDups",
                  "BND": " Breakends (TRA/TPO)"}
    label = []
    data = []
    for key in svdict:
        if svdict[key] > 0:
            label.append(str(svdict[key]) + svnamedict[key])
            data.append(svdict[key])

    def func(pct, allvals):
        absolute = int(pct/100.*np.sum(allvals))
        return "{:.1f}%".format(pct, absolute)

    explode = []
    for i in range(len(data)):
        explode.append(0.02)
    explode = tuple(explode)  # explode = (0.02, 0.02, 0.02, 0.02, 0.02)
    wedges, text, autotexts = ax.pie(data, wedgeprops=dict(width=0.5), startangle=-40, autopct=lambda pct: func(pct, data),
                                     textprops=dict(color="white"), counterclock=False, pctdistance=0.75,
                                     colors=['#7d3f5d', '#7d5f3f', '#3f5d7d', '#3f7c7d', '#403f7d'],
                                     explode=explode)
    bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.72)
    kw = dict(xycoords='data', textcoords='data', arrowprops=dict(arrowstyle="-"),
              bbox=bbox_props, zorder=0, va="center")

    for i, p in enumerate(wedges):
        ang = (p.theta2 - p.theta1)/2. + p.theta1
        y = np.sin(np.deg2rad(ang))
        x = np.cos(np.deg2rad(ang))
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle})
        if data[i] > 0:
            ax.annotate(label[i], xy=(x, y), xytext=(1.35*np.sign(x), 1.4*y),
                        horizontalalignment=horizontalalignment, **kw)

    plt.setp(autotexts, size=11)
    ax.text(0, 0, 'SV types', ha='center', fontsize=15)
    plt.savefig(os.path.join(fwd, 'sv_type_donut.png'), bbox_inches='tight', dpi=100, facecolor=fig.get_facecolor(),
                edgecolor='none')


# Create read length distribution plot
def read_len_dict(rlen_dict, fwd):
    qlen = measureqlen(rlen_dict)
    m = math.ceil(max(np.log10(np.array(qlen))))
    e = math.floor(min(np.log10(np.array(qlen))))
    bins = 10**np.linspace(e, m, m*20)
    fig = plt.figure(figsize=(8, 6))
    fig.patch.set_facecolor('#f6f7f9')
    ax = fig.add_subplot(111)
    ax.hist(qlen, bins=bins, color="#403f7d", edgecolor="#403f7d", linewidth=0.5, histtype='stepfilled')
    ax.set_xscale('log')
    ax.xaxis.set_major_formatter(ScalarFormatter())
    ax.ticklabel_format(useOffset=False, style='plain')
    ax.set_facecolor('#ebebff')
    plt.xlabel("Read length (bases)")
    plt.ylabel("Number of reads")
    plt.savefig(os.path.join(fwd, 'read_length_dist.png'), bbox_inches='tight', dpi=100, facecolor=fig.get_facecolor(),
                edgecolor='none')


# Get read length distribution
def measureqlen(rlen_dict):
    qlen = []
    for key in rlen_dict:
        qlen.append(rlen_dict[key])
    return qlen


# Create html
def create_html(data, fwd, wk_dir, vcf_path, timenow, read_name, read_path, ref_path, threshold, n, totalsv, css, js):
    html = open(os.path.join(wk_dir, '%s.nanovar.pass.report-tmp.html' % read_name), 'w')
    begin = """<!DOCTYPE html>
    <html lang="en">
    <head>
        <meta charset="utf-8">
        <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
        <meta http-equiv="x-ua-compatible" content="ie=edge">
        <title>NanoVar Report</title>
        <!-- Font Awesome -->
        <link href="{}/font-awesome.min.css" rel="stylesheet">
        <!-- Bootstrap core CSS -->
        <link href="{}/bootstrap.min.css" rel="stylesheet">
        <!-- Material Design Bootstrap -->
        <link href="{}/mdb.min.css" rel="stylesheet">
        <!-- DataTable CSS -->
        <link href="{}/jquery.dataTables.min.css" rel="stylesheet">
        <!-- DataTable buttons CSS -->
        <link href="{}/buttons.dataTables.min.css" rel="stylesheet">
        """.format(css, css, css, css, css) + """
        <style>
            body {
                background-color: #f6f7f9;
            }
            h1 { 
                margin-left: 4px;
                font-weight: bold;
                font-size: 40px;
                font-family: sans-serif;
            }
            h2, h3, h4, h5, p { 
                margin-left: 4px;
                font-family: sans-serif;
            }
            table {
                font-family: sans-serif;
            }
            #image {
                text-align:center;
            }
            #margin {
                margin-left: 35px;
                margin-right: 35px;
                margin-top: 25px;
            }
            .floatleft {
                float: left;
                width: 45%;
                margin-right: 5%;
                padding-top: 80px
            }
            .floatright {
                float: right;
                width: 45%;
                margin-left: 5%;
            }
            .container {
                overflow;
                max-width: 1400px;
            }
            .aligncenter {
                float: center;
            }
            .alignleft {
                float: left;
            }
            .alignright {
                float: right;
            }
        </style>
    </head>
    <body>
    <div id="margin">
        <div>
            <h1 style="text-align:center;">NanoVar Report</h1>
            <h5 style="text-align:center;">Version: NanoVar """ + __version__ + """</h5>
            <h5 style="text-align:center;">""" + str(timenow.strftime("%a %d %B %Y")) + """</h5>
            <h5 style="text-align:center;">""" + read_name + """</h5>
            <br>
            <br>
            <div style="background-color:#3f5d7d;color:#f6f7f9">
                <h4 style="text-align:center;">Run details</h4>
            </div>
        </div>
        <div style="clear: both;"></div>
        <br>
        <table width="50%" class="table-sm table-bordered" style="table-layout:fixed" align="center">
            <col width="150">
            <thead>
            <tr>
                <th style="background-color:#2e6096; color:#f6f7f9; font-weight: bold; text-align:center;">Variable</th>
                <th style="background-color:#2e6096; color:#f6f7f9; font-weight: bold; text-align:center;">Value</th>
            </tr>
            </thead>
            <tbody>
            <tr>
                <td>Input reads:</td>
                <td style="word-wrap:break-word;">""" + read_path + """</td>
            </tr>
            <tr>
                <td>Reference genome:</td>
                <td style="word-wrap:break-word;">""" + ref_path + """</td>
            </tr>
            <tr>
                <td>Output VCF:</td>
                <td style="word-wrap:break-word;">""" + vcf_path + """</td>
            </tr>
            <tr>
                <td>SV score threshold:</td>
                <td>""" + str(threshold) + """</td>
            </tr>
            </tbody>
        </table>
        <br>
        <br>
        <br>
        <br>
        <div style="background-color:#366b6c;color:#f6f7f9;">
            <h3 style="text-align:center;">Results</h3>
        </div>
        <p style="clear: both;">
        <br>
        <h4 style="text-align:center;"><u>1. Table of output SVs</u></h4>
        <h5 style="text-align:center;">Showing """ + str(len(data)) + """ out of """ + str(totalsv) + """ total SVs</h5>
        <table id="NanoVar_report_table" class="table table-striped table-bordered table-sm" cellspacing="0" width="100%">
            <thead>
                <tr>
                    <th class="th-sm">#
                        <i class="fa fa-sort float-right" aria-hidden="true"></i>
                    </th>
                    <th class="th-sm">SV type
                        <i class="fa fa-sort float-right" aria-hidden="true"></i>
                    </th>
                    <th class="th-sm">Chrom1
                        <i class="fa fa-sort float-right" aria-hidden="true"></i>
                    </th>
                    <th class="th-sm">Pos1
                        <i class="fa fa-sort float-right" aria-hidden="true"></i>
                    </th>
                    <th class="th-sm">Chrom2
                        <i class="fa fa-sort float-right" aria-hidden="true"></i>
                    </th>
                    <th class="th-sm">Pos2
                        <i class="fa fa-sort float-right" aria-hidden="true"></i>
                    </th>
                    <th class="th-sm">Length
                        <i class="fa fa-sort float-right" aria-hidden="true"></i>
                    </th>
                    <th class="th-sm">Score
                        <i class="fa fa-sort float-right" aria-hidden="true"></i>
                    </th>
                    <th class="th-sm">No. of breakend-supporting reads
                        <i class="fa fa-sort float-right" aria-hidden="true"></i>
                    </th>
                    <th class="th-sm">No. of breakend-opposing reads
                        <i class="fa fa-sort float-right" aria-hidden="true"></i>
                    </th>
                    <th title="No. of breakend-supporting reads/total reads at breakend" class="th-sm">Breakend read ratio
                        <i class="fa fa-sort float-right" aria-hidden="true"></i>
                    </th>
                    <th class="th-sm">Approximated genotype
                        <i class="fa fa-sort float-right" aria-hidden="true"></i>
                    </th>
                    <th title="Last five digits define the breakend ID code" class="th-sm">SV ID
                        <i class="fa fa-sort float-right" aria-hidden="true"></i>
                    </th>
                </tr>
            </thead>
            <tbody>
    """
    html.write(begin)

    # Create table body
    row = ""
    for i in data:
        row += """          <tr>
    """
        for j in i:
            row += "                <td>" + j + """</td>
    """
        row += """          </tr>
    """
    row += """      </tbody>
        </table>
        <br>
        <br>
        <div id="image">
            <figure>
                <h4 style="text-align:center;"><u>2. Distribution of SV types</u></h4>
                <img src=""" + '"' + fwd + """/sv_type_donut.png" alt="2. Distribution of SV types" title=""" + '"' + fwd + \
           '/sv_type_donut.png' + '"' + """>
            </figure>
            <br>
            <br>
            <figure>
                <h4 style="text-align:center;"><u>3. Size distribution of SVs across SV types</u></h4>
                <img src=""" + '"' + fwd + """/sv_lengths.png" alt="3. Size distribution of SVs across SV types" title=""" + \
           '"' + fwd + '/sv_lengths.png' + '"' + """>
            </figure>
            <br>
            <br>
            <figure>
                <h4 style="text-align:center;"><u>4. Scatter plot between SV confidence score and read ratio</u></h4>
                <img src=""" + '"' + fwd + """/scatter1.png" alt="4. Scatter plot between SV confidence score and read ratio" 
                title=""" + '"' + fwd + '/scatter1.png' + '"' + """>
            </figure>
            <br>
            <br>
            <figure>
                <h4 style="text-align:center;"><u>5. Scatter plot between SV confidence score and supporting read depth</u></h4>
                <img src=""" + '"' + fwd + """/scatter2.png" alt="5. Scatter plot between SV confidence score and supporting read depth" 
                title=""" + '"' + fwd + '/scatter2.png' + '"' + """>
            </figure>
            <br>
            <br>
            <div style="background-color:#403f7d;color:#f6f7f9;">
                <h3 style="text-align:center;">QC Statistics</h3>
            </div>
            <br>
            <figure>
                <h4 style="text-align:center;"><u>6. Distribution of read lengths of input FASTA</u></h4>
                <img src=""" + '"' + fwd + """/read_length_dist.png" alt="6. Distribution of read lengths of input FASTA" 
                title=""" + '"' + fwd + '/read_length_dist.png' + '"' + """>
            </figure>
            <br>
            <br>
            <figure>
                <h4 style="text-align:center;"><u>7. Depth of coverage across genome</u></h4>
                <img src=""" + '"' + fwd + """/depth_of_coverage.png" alt="7. Depth of coverage across genome" title=""" + '"' \
           + fwd + '/figures/depth_of_coverage.png' + '"' + """>
            </figure>
            <br>
        </div>
        <br>
        <br>
        <!-- SCRIPTS -->
        <!-- JQuery -->
        <script type="text/javascript" src="{}/jquery-3.3.1.min.js"></script>
        <!-- datatable javascript  -->
        <script type="text/javascript" src="{}/jquery.dataTables.min.js"></script>
        <!-- datatable buttons javascript  -->
        <script type="text/javascript" src="{}/dataTables.buttons.min.js"></script>
        <!-- buttons flash javascript  -->
        <script type="text/javascript" src="{}/buttons.flash.min.js"></script>
        <!-- buttons jszip javascript  -->
        <script type="text/javascript" src="{}/jszip.min.js"></script>
        <!-- buttons html5 buttons javascript  -->
        <script type="text/javascript" src="{}/buttons.html5.min.js"></script>
        <!-- Bootstrap tooltips -->
        <script type="text/javascript" src="{}/popper.min.js"></script>
        <!-- Bootstrap core JavaScript -->
        <script type="text/javascript" src="{}/bootstrap.min.js"></script>
        <!-- MDB core JavaScript -->
        <script type="text/javascript" src="{}/mdb.min.js"></script>
        """.format(js, js, js, js, js, js, js, js, js) + """
        <script type="text/javascript">
            $('#NanoVar_report_table').DataTable({
                "scrollX": true,
                "scrollY": 200,
                dom: 'Bfrtip',
                buttons: [
                    'copy', 'csv', 'excel'
                ]
            });
            $('.dataTables_length').addClass('bs-select');
        </script>
    </div>
    </body>
    </html>
    """
    html.write(row)
    html.close()
    text = open(os.path.join(wk_dir, '%s.nanovar.pass.report-tmp.html' % read_name)).read()
    soup = BeautifulSoup(text, "html.parser")
    tag_list = []
    tag_list += soup('img')
    css_tags = soup('link')
    for css in css_tags:
        if 'stylesheet' in css['rel']:
            tag_list.append(css)
    js_tags = soup('script')
    for script in js_tags:
        if 'src' in script.attrs:
            tag_list.append(script)
    for tag in tag_list:
        tag_url = tag['href'] if tag.name == 'link' else tag['src']
        with open(tag_url, 'rb') as f:
            data = f.read()
        mtype, _ = mimetypes.guess_type(tag_url)
        if tag.name == 'link':
            en_data = quote(data.decode())
            tag['href'] = "data:{},{}".format(mtype, en_data)
        elif tag.name == 'script':
            en_data = quote(data.decode())
            tag['src'] = "data:{},{}".format(mtype, en_data)
        elif tag.name == 'img':
            mtype = mtype + ';base64'
            en_data = base64.b64encode(data).decode()
            tag['src'] = "data:{},{}".format(mtype, en_data)
    # packed_html = str(soup)
    with open(os.path.join(wk_dir, f"{read_name}.nanovar.pass.report.html"), "w", encoding = 'utf-8') as file: 
        file.write(str(soup.prettify()))
    # packed_html = htmlark.convert_page(os.path.join(wk_dir, '%s.nanovar.pass.report-tmp.html' % read_name), ignore_errors=True)
    # html_final = open(os.path.join(wk_dir, '%s.nanovar.pass.report.html' % read_name), 'w')
    # _ = html_final.write(packed_html)
    # html_final.close()
    os.remove(os.path.join(wk_dir, '%s.nanovar.pass.report-tmp.html' % read_name))



"""
Transposable element analyzer.

Copyright (C) 2021 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""


import os
import logging
from pybedtools import BedTool
from subprocess import Popen, PIPE, STDOUT


def te_analyzer(wk_dir, out_nn, total_out1, total_out2, score_threshold, ref_dir, hsb, mdb, threads, debug):
    # Convert score threshold to nn probability
    nn_score = 1-10**(score_threshold/-10)
    index2reads = {}
    for i in out_nn:
        if i.split('\t')[3].split(' ')[0] in ['Nov_Ins', 'S-Nov_Ins_bp', 'E-Nov_Ins_bp']:
            if float(i.split('\t')[13]) >= nn_score:
                ind = i.split('\t')[6].split('~')[0]
                reads = i.split('\t')[8] + ',' + i.split('\t')[11]
                index2reads[ind] = reads
    region_dict = {}
    for i in total_out1:
        if i.split('\t')[3][:3] == 'S-N':
            region_dict[i.split('\t')[8]] = (1, float(i.split('\t')[3].split(' ')[1].split('~')[0]))
        elif i.split('\t')[3][:3] == 'E-N':
            region_dict[i.split('\t')[8]] = (int(i.split('\t')[1]), int(i.split('\t')[1]) +
                                             float(i.split('\t')[3].split(' ')[1].split('~')[0]))
        elif i.split('\t')[3][:7] == 'Nov_Ins':
            region_dict[i.split('\t')[8]] = (int(i.split('\t')[1]), int(i.split('\t')[1]) +
                                             float(i.split('\t')[3].split(' ')[1].split('~')[0]))
        else:
            continue
    for i in total_out2:
        if i.split('\t')[3][:3] == 'S-N':
            region_dict[i.split('\t')[8]] = (1, float(i.split('\t')[3].split(' ')[1].split('~')[0]))
        elif i.split('\t')[3][:3] == 'E-N':
            region_dict[i.split('\t')[8]] = (int(i.split('\t')[1]), int(i.split('\t')[1]) +
                                             float(i.split('\t')[3].split(' ')[1].split('~')[0]))
        elif i.split('\t')[3][:7] == 'Nov_Ins':
            region_dict[i.split('\t')[8]] = (int(i.split('\t')[1]), int(i.split('\t')[1]) +
                                             float(i.split('\t')[3].split(' ')[1].split('~')[0]))
        else:
            continue
    bed_str = ''
    read2index = {}
    for i in index2reads:
        for r in index2reads[i].split(','):
            try:
                left = str(region_dict[r.strip('\n')][0])
                right = str(int(region_dict[r.strip('\n')][1]))
                if r.split('~')[0] + ':' + left + '-' + right not in read2index:
                    bed_str += r.split('~')[0] + '\t' + left + '\t' + right + '\n'
                    read2index[r.split('~')[0] + ':' + left + '-' + right] = i
            except KeyError:
                pass
    bed = BedTool(bed_str, from_string=True)
    # Extract fasta for each insertion region
    fasta_path = os.path.join(wk_dir, 'temp1.fa')
    fasta = bed.sequence(fi=fasta_path)
    _ = fasta.save_seqs(os.path.join(wk_dir, 'ins_seq.fa'))
    # HS-BLASTN alignment
    ref = os.path.join(ref_dir, 'hg38_L1_Alu.fa')
    # Index TE reference (makeblastdb)
    process = Popen([mdb, '-in', ref, '-input_type', 'fasta', '-dbtype', 'nucl'], universal_newlines=True,
                    stdout=PIPE, stderr=STDOUT)
    with process.stdout:
        log_subprocess(process.stdout)
    exitcode = process.wait()
    if exitcode != 0:
        logging.critical("Error: makeblastdb failed (2)")
        raise Exception("Error: makeblastdb failed (2), see log")
    # Index TE reference (hs-blastn)
    process = Popen([hsb, 'index', ref], universal_newlines=True, stdout=PIPE, stderr=STDOUT)
    with process.stdout:
        log_subprocess(process.stdout)
    exitcode = process.wait()
    if exitcode != 0:
        logging.critical("Error: hs-blastn index failed (2)")
        raise Exception("Error: hs-blastn index failed (2), see log")
    obinary_path = os.path.join(ref_dir, 'hg38_L1_Alu.counts.obinary')
    read_path = os.path.join(wk_dir, 'ins_seq.fa')
    out_path = os.path.join(wk_dir, 'temp-te-blast.tab')
    process = Popen([hsb + ' align -db ' + ref + ' -window_masker_db ' + obinary_path + ' -query ' + read_path + ' -out '
                     + out_path + ' -outfmt 6 -num_threads ' + str(threads) + ' -max_target_seqs 3 -gapopen 0 '
                    '-gapextend 4 -penalty -3 -reward 2'],
                    universal_newlines=True, stdout=PIPE, stderr=STDOUT, shell=True, executable='/bin/bash')
    with process.stdout:
        log_subprocess(process.stdout)
    exitcode = process.wait()
    if exitcode != 0:
        logging.critical("Error: hs-blastn alignment failed (2)")
        raise Exception("Error: hs-blastn alignment failed (2), see log")
    index2te = {}
    with open(out_path) as f:
        for line in f:
            line = line.split('\t')
            if line[1].rsplit('_', 1)[1].startswith('L1'):
                if int(line[3]) >= 5500:  # 90% of L1
                    if read2index[line[0]] in index2te:
                        index2te[read2index[line[0]]].add(line[1].rsplit('_', 1)[1])
                    else:
                        index2te[read2index[line[0]]] = set()
                        index2te[read2index[line[0]]].add(line[1].rsplit('_', 1)[1])
            elif line[1].rsplit('_', 1)[1].startswith('Alu'):
                if int(line[3]) >= 270:  # 90% of Alu
                    if read2index[line[0]] in index2te:
                        index2te[read2index[line[0]]].add(line[1].rsplit('_', 1)[1])
                    else:
                        index2te[read2index[line[0]]] = set()
                        index2te[read2index[line[0]]].add(line[1].rsplit('_', 1)[1])
    if not debug:
        os.remove(read_path)
        os.remove(out_path)
        
    return index2te


def log_subprocess(out):
    for line in iter(out.readline, ''):
        if line != '\n':
            logging.debug(line.strip())



"""
Parsing algorithm.

Copyright (C) 2021 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""

import random
import re
import ast
import uuid


# Parse alignment entries
def entry_parser(temp1, chromocollect, overlap_tol):
    temp2 = []
    num_chr = len(chromocollect)  # Total number of chromosomes aligned
    num_align = len(temp1)  # Total number of alignments
    while len(temp1) != 0:
        # Lock-in lead alignment with the highest bitscore
        temp2.append(temp1[0] + '\tn=' + str(num_align) + '\t' + str(num_chr) + 'chr')
        # leadrange = [int(temp1[0].split('\t')[5]), int(temp1[0].split('\t')[5]) + int(temp1[0].split('\t')[6])]
        leadstart = int(temp1[0].split('\t')[5])
        leadend = int(temp1[0].split('\t')[5]) + int(temp1[0].split('\t')[6])
        del temp1[0]  # Remove lead alignment entry
        j = len(temp1)  # Recalculate number of remaining entries
        temp3 = []
        for p in range(j):
            qstart = int(temp1[p].split('\t')[5])
            qend = int(temp1[p].split('\t')[5]) + int(temp1[p].split('\t')[6])
            queryrange = range(qstart, qend + 1)
            leadintersect = range(max(leadstart, qstart), min(leadend, qend) + 1)
            if len(leadintersect) == 0:  # No intersection with lead
                temp3.append(temp1[p])
            else:
                qulen = len(queryrange)
                leadintlen = len(leadintersect)
                if float(leadintlen) / qulen > overlap_tol or (qulen - leadintlen) <= 50:
                    # If overlap len is more than 90% of query alignment length or
                    # after-trimmed length <=50 bp, lead wins and query is omitted. Otherwise, do trimming.
                    continue  # Entry omitted
                else:  # If overlap is tolerated, then do trimming
                    new = ''
                    strand = str(temp1[p].split('\t')[7])
                    # Alter query and subject ranges and adjust bitscore as a proportion of alignment length
                    if qstart == min(leadintersect):  # Left overlap
                        if strand == '+':
                            new = temp1[p].split('\t')[0] + '\t' + str(
                                int(temp1[p].split('\t')[1]) + leadintlen) + '\t' + str(
                                int(temp1[p].split('\t')[2]) - leadintlen) + '\t' + '\t'.join(
                                temp1[p].split('\t')[3:5]) + '\t' + str(
                                int(temp1[p].split('\t')[5]) + leadintlen) + '\t' + str(
                                int(temp1[p].split('\t')[6]) - leadintlen) + '\t' + '\t'.join(
                                temp1[p].split('\t')[7:10]) + '\t' + str(
                                round(float(temp1[p].split('\t')[10]) * (float(qulen - leadintlen) / qulen),
                                      2)) + '\t' + '\t'.join(temp1[p].split('\t')[11:])
                        elif strand == '-':
                            new = temp1[p].split('\t')[0] + '\t' + temp1[p].split('\t')[1] + '\t' + str(
                                int(temp1[p].split('\t')[2]) - leadintlen) + '\t' + '\t'.join(
                                temp1[p].split('\t')[3:5]) + '\t' + str(
                                int(temp1[p].split('\t')[5]) + leadintlen) + '\t' + str(
                                int(temp1[p].split('\t')[6]) - leadintlen) + '\t' + '\t'.join(
                                temp1[p].split('\t')[7:10]) + '\t' + str(
                                round(float(temp1[p].split('\t')[10]) * (float(qulen - leadintlen) / qulen),
                                      2)) + '\t' + '\t'.join(temp1[p].split('\t')[11:])
                    elif qend == max(leadintersect):  # Right overlap
                        if strand == '+':
                            new = temp1[p].split('\t')[0] + '\t' + temp1[p].split('\t')[1] + '\t' + str(
                                int(temp1[p].split('\t')[2]) - leadintlen) + '\t' + '\t'.join(
                                temp1[p].split('\t')[3:5]) + '\t' + temp1[p].split('\t')[5] + '\t' + str(
                                int(temp1[p].split('\t')[6]) - leadintlen) + '\t' + '\t'.join(
                                temp1[p].split('\t')[7:10]) + '\t' + str(
                                round(float(temp1[p].split('\t')[10]) * (float(qulen - leadintlen) / qulen),
                                      2)) + '\t' + '\t'.join(temp1[p].split('\t')[11:])
                        elif strand == '-':
                            new = temp1[p].split('\t')[0] + '\t' + str(
                                int(temp1[p].split('\t')[1]) + leadintlen) + '\t' + str(
                                int(temp1[p].split('\t')[2]) - leadintlen) + '\t' + '\t'.join(
                                temp1[p].split('\t')[3:5]) + '\t' + temp1[p].split('\t')[5] + '\t' + str(
                                int(temp1[p].split('\t')[6]) - leadintlen) + '\t' + '\t'.join(
                                temp1[p].split('\t')[7:10]) + '\t' + str(
                                round(float(temp1[p].split('\t')[10]) * (float(qulen - leadintlen) / qulen),
                                      2)) + '\t' + '\t'.join(temp1[p].split('\t')[11:])
                    else:  # If lead lies entirely within intersection, omit entry
                        continue
                    if new != '' and int(new.split('\t')[2]) > 0:
                        temp3.append(new)
        temp1 = temp3
    sortdict = {}
    for k in temp2:
        sortdict[int(k.split('\t')[5])] = k
    return [value for (key, value) in sorted(sortdict.items())]


# Extract alignment info
def align_info(line, rlendict):
    qid = line.split('\t')[0]
    sid = line.split('\t')[1]
    piden = line.split('\t')[2]
    nmismatch = line.split('\t')[4]
    ngapopen = line.split('\t')[5]
    qstart = line.split('\t')[6]
    qstretch = int(line.split('\t')[7]) - int(line.split('\t')[6])
    sstart = min(int(line.split('\t')[8]), int(line.split('\t')[9]))
    sstretch = abs(int(line.split('\t')[9]) - int(line.split('\t')[8]))
    evalue = line.split('\t')[10]
    bitscore = line.split('\t')[11].strip('\n')
    if line.split('\t')[8] <= line.split('\t')[9]:
        strand = "+"
    else:
        strand = "-"
    rlen = rlendict[qid]
    entry = sid + '\t' + str(sstart) + '\t' + str(sstretch) + '\t+\t' + qid + '\t' + qstart + '\t' + str(qstretch) + '\t' + \
        strand + '\t' + str(rlen) + '\t' + evalue + '\t' + bitscore + '\t' + piden + '\t' + nmismatch + '\t' + ngapopen
    return entry


# Parse SV breakpoints
def breakpoint_parser(out, minlen, sig_index, seed, aligner):
    rd = random.Random()
    rd.seed(seed)
    uname_repeat = {}
    final = []
    if aligner == 'mm':
        ran_len = 5
    elif aligner == 'hsb':
        ran_len = 4
    else:
        raise Exception('ERROR: Internal error - breakpoint_parser aligner parameter fault')
    # ran = "01234567890ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    # random.seed(seed)
    read_name = out.split('\t')[0]
    if out.split('\t')[17] == '':
        complex_sv = 0
    else:
        complex_sv = len(out.split('\t')[17].split(','))
    # if int(out.split('\t')[17].count(',')) >= 1:
    #     multi_sv = len(out.split('\t')[18].split(','))
    # else:
    #     multi_sv = 0
    del_count = del_counter = int(out.split('\t')[19].count('Del'))
    iso_ins_count = iso_ins_counter = int(out.split('\t')[19].count('S-Nov_Ins')) + int(out.split('\t')[19].count('E-Nov_Ins'))
    ins_count = ins_counter = int(out.split('\t')[19].count('Nov_Ins'))
    tdup_count = tdup_counter = int(out.split('\t')[19].count('TDupl'))
    inv_count = inv_counter = int(out.split('\t')[19].count('Inv'))
    intra_ins_count = intra_ins_counter = int(out.split('\t')[19].count('Intra-Ins'))
    inter_ins_count = inter_ins_counter = int(out.split('\t')[19].count('Inter-Ins'))
    inter_tx_count = inter_tx_counter = int(out.split('\t')[19].count('InterTx'))
    strandness = out.split('\t')[15].split(',')
    bps = out.split('\t')[19].split(',')
    bps_no = len(bps)
    querymap = out.split('\t')[20]
    nchrom = out.split('\t')[1]
    evaluelist = out.split('\t')[4].split(',')
    bitscorelist = out.split('\t')[5].split(',')
    short = out.split('\t')[16]
    piden = out.split('\t')[21].split(',')
    mismatch = out.split('\t')[22].split(',')
    gap_ratio = out.split('\t')[23].split(',')
    sign = out.split('\t')[3]
    realnmaps = int(out.split('\t')[2].split(' ')[0])
    # if realnmaps >= 5:
    #     final = []
    #     return final
    for j in range(bps_no):
        bp_name = bps[j].split(':')[0]
        bp_range = re.sub('^.*:', '', str(bps[j]))
        bp1 = int(bp_range.split('-')[0])
        bp2 = int(bp_range.split('-')[1])
        signature, strands = getsignature(bp1, bp_name, querymap, sign, realnmaps, evaluelist, bitscorelist, bps_no,
                                          complex_sv, nchrom, short, piden, mismatch, gap_ratio, strandness, sig_index)
        bp_uname = uuid.UUID(int=rd.getrandbits(128), version=4).hex[:ran_len].upper()
        # bp_uname = ''.join(random.choices(ran, k=ran_len))
        try:
            while uname_repeat[bp_uname]:
                bp_uname = uuid.UUID(int=rd.getrandbits(128), version=4).hex[:ran_len].upper()
        except KeyError:
            uname_repeat[bp_uname] = 1
        if bp_name == 'S-Nov_Ins':
            chrom = out.split('\t')[7].split(' ')[1].split(',')[int(iso_ins_count - iso_ins_counter)].split(':')[0]
            coord = out.split('\t')[7].split(' ')[1].split(',')[int(iso_ins_count - iso_ins_counter)].split(':')[1]
            s_nov_ins_size = out.split('\t')[7].split(' ')[0].split(',')[int(iso_ins_count - iso_ins_counter)].split('ns')[1]
            pair = '.'
            uniqidx = bp_uname + '~' + str(chrom) + ':' + str(coord)
            uniqname = read_name + '~' + bp_uname
            if int(s_nov_ins_size) >= minlen:
                final.append(
                    str(read_name) + '\t' + str(bp2 - 10) + '\t' + str(bp2) + '\t' + str(bp_name + '_bp') + ' ' +
                    str(s_nov_ins_size) + '~' + str(strands) + '\t' + str(chrom) + '\t' + str('-') + '\t' +
                    str(uniqidx) + '\t' + str(pair) + '\t' + str(uniqname) + '\t' + str(signature)
                )
            iso_ins_counter = iso_ins_counter - 1
        elif bp_name == 'E-Nov_Ins':
            chrom = out.split('\t')[7].split(' ')[1].split(',')[int(iso_ins_count - iso_ins_counter)].split(':')[0]
            coord = out.split('\t')[7].split(' ')[1].split(',')[int(iso_ins_count - iso_ins_counter)].split(':')[1]
            e_nov_ins_size = out.split('\t')[7].split(' ')[0].split(',')[int(iso_ins_count - iso_ins_counter)].split('ns')[1]
            pair = '.'
            uniqidx = bp_uname + '~' + str(chrom) + ':' + str(coord)
            uniqname = read_name + '~' + bp_uname
            if int(e_nov_ins_size) >= minlen:
                final.append(
                    str(read_name) + '\t' + str(bp1) + '\t' + str(bp1 + 10) + '\t' + str(bp_name + '_bp') + ' ' +
                    str(e_nov_ins_size) + '~' + str(strands) + '\t' + str(chrom) + '\t' + str('-') + '\t' +
                    str(uniqidx) + '\t' + str(pair) + '\t' + str(uniqname) + '\t' + str(signature)
                )
            iso_ins_counter = iso_ins_counter - 1
        elif bp_name == 'Del':
            chrom = out.split('\t')[6].split(' ')[1].split(',')[int(del_count - del_counter)].split(':')[0]
            coord1 = int(out.split('\t')[6].split(' ')[1].split(',')[int(del_count - del_counter)].split(':')[1].split('-')[0])
            coord2 = int(out.split('\t')[6].split(' ')[1].split(',')[int(del_count - del_counter)].split(':')[1].split('-')[1])
            del_size = out.split('\t')[6].split(' ')[0].split(',')[int(del_count - del_counter)].split('l')[1]
            pair = '.'
            uniqidx = bp_uname + '~' + str(chrom) + ':' + str(min(coord1, coord2)) + '-' + str(max(coord1, coord2))
            uniqname = read_name + '~' + bp_uname
            if int(del_size) >= minlen:
                final.append(
                    str(read_name) + '\t' + str(bp1) + '\t' + str(bp2) + '\t' + str(bp_name) + ' ' + str(del_size) + '~' +
                    str(strands) + '\t' + str(chrom) + '\t' + str('-') + '\t' + str(uniqidx) + '\t' + str(pair) + '\t' +
                    str(uniqname) + '\t' + str(signature)
                )
            del_counter = del_counter - 1
        elif bp_name == 'Nov_Ins':
            chrom = out.split('\t')[8].split(' ')[1].split(',')[int(ins_count - ins_counter)].split(':')[0]
            coord1 = int(out.split('\t')[8].split(' ')[1].split(',')[int(ins_count - ins_counter)].split(':')[1].split('-')[0])
            coord2 = int(out.split('\t')[8].split(' ')[1].split(',')[int(ins_count - ins_counter)].split(':')[1].split('-')[1])
            nov_ins_size = out.split('\t')[8].split(' ')[0].split(',')[int(ins_count - ins_counter)].split('ns')[1]
            uniqidx = bp_uname + '~' + str(chrom) + ':' + str(min(coord1, coord2)) + '-' + str(max(coord1, coord2))
            uniqname = read_name + '~' + bp_uname
            if int(nov_ins_size) >= minlen:
                final.append(
                    str(read_name) + '\t' + str(bp1) + '\t' + str(bp1 + 10) + '\t' + str(bp_name) + ' ' +
                    str(nov_ins_size) + '~' + str(strands) + '\t' + str(chrom) + '\t' + str('-') + '\t' +
                    str(uniqidx) + '\t' + str('Nov1') + '\t' + str(uniqname) + '\t' + str(signature)
                )
            ins_counter = ins_counter - 1
        elif bp_name == 'TDupl':
            chrom = out.split('\t')[10].split(' ')[1].split(',')[int(tdup_count - tdup_counter)].split(':')[0]
            coord1 = int(out.split('\t')[10].split(' ')[1].split(',')[int(tdup_count - tdup_counter)].split(':')[1].split('-')[0])
            coord2 = int(out.split('\t')[10].split(' ')[1].split(',')[int(tdup_count - tdup_counter)].split(':')[1].split('-')[1])
            pair = '.'
            uniqidx = bp_uname + '~' + str(chrom) + ':' + str(min(coord1, coord2)) + '-' + str(max(coord1, coord2))
            uniqname = read_name + '~' + bp_uname
            final.append(
                str(read_name) + '\t' + str(bp1) + '\t' + str(bp2) + '\t' + str(bp_name) + ' 99.99' + '~' +
                str(strands) + '\t' + str(chrom) + '\t' + str('-') + '\t' + str(uniqidx) + '\t' + str(pair) + '\t' +
                str(uniqname) + '\t' + str(signature)
            )
            tdup_counter = tdup_counter - 1
        elif bp_name == 'Inv':
            chrom = out.split('\t')[11].split(' ')[1].split(',')[int(inv_count - inv_counter)].split(':')[0]
            coord1 = int(out.split('\t')[11].split(' ')[1].split(',')[int(inv_count - inv_counter)].split(':')[1].split('-')[0])
            coord2 = int(out.split('\t')[11].split(' ')[1].split(',')[int(inv_count - inv_counter)].split(':')[1].split('-')[1])
            pair = '.'
            uniqidx = bp_uname + '~' + str(chrom) + ':' + str(coord1) + '-' + str(coord2)
            uniqname = read_name + '~' + bp_uname
            final.append(
                str(read_name) + '\t' + str(bp1) + '\t' + str(bp2) + '\t' + str(bp_name) + ' 99.99' + '~' +
                str(strands) + '\t' + str(chrom) + '\t' + str('-') + '\t' + str(uniqidx) + '\t' + str(pair) + '\t' +
                str(uniqname) + '\t' + str(signature)
            )
            inv_counter = inv_counter - 1
        elif bp_name == 'Inv(1)':
            chrom = out.split('\t')[11].split(' ')[1].split(',')[int(inv_count - inv_counter)].split(':')[0]
            coord1 = int(out.split('\t')[11].split(' ')[1].split(',')[int(inv_count - inv_counter)].split(':')[1].split('-')[0])
            coord2 = int(out.split('\t')[11].split(' ')[1].split(',')[int(inv_count - inv_counter)].split(':')[1].split('-')[1])
            inv_size = out.split('\t')[11].split(' ')[0].split(',')[int(inv_count - inv_counter)].split('nv')[1]
            pair = 'Inv1'
            uniqidx = bp_uname + '~' + str(chrom) + ':' + str(coord1) + '-' + str(coord2)
            uniqname = read_name + '~' + bp_uname
            if int(inv_size) >= minlen:
                final.append(
                    str(read_name) + '\t' + str(bp1) + '\t' + str(bp2) + '\t' + str(bp_name) + ' ' + str(inv_size) + '~' +
                    str(strands) + '\t' + str(chrom) + '\t' + str('-') + '\t' + str(uniqidx) + '\t' + str(pair) + '\t' +
                    str(uniqname) + '\t' + str(signature)
                )
            inv_counter = inv_counter - 1
        elif bp_name == 'Inv(2)':
            chrom = out.split('\t')[11].split(' ')[1].split(',')[int(inv_count - inv_counter)].split(':')[0]
            coord1 = int(out.split('\t')[11].split(' ')[1].split(',')[int(inv_count - inv_counter)].split(':')[1].split('-')[0])
            coord2 = int(out.split('\t')[11].split(' ')[1].split(',')[int(inv_count - inv_counter)].split(':')[1].split('-')[1])
            inv_size = out.split('\t')[11].split(' ')[0].split(',')[int(inv_count - inv_counter)].split('nv')[1]
            pair = 'Inv2'
            uniqidx = bp_uname + '~' + str(chrom) + ':' + str(coord1) + '-' + str(coord2)
            uniqname = read_name + '~' + bp_uname
            if int(inv_size) >= minlen:
                final.append(
                    str(read_name) + '\t' + str(bp1) + '\t' + str(bp2) + '\t' + str(bp_name) + ' ' + str(inv_size) + '~' +
                    str(strands) + '\t' + str(chrom) + '\t' + str('-') + '\t' + str(uniqidx) + '\t' + str(pair) + '\t' +
                    str(uniqname) + '\t' + str(signature)
                )
            inv_counter = inv_counter - 1
        elif bp_name == 'Intra-Ins':
            chrom = out.split('\t')[12].split(' ')[1].split(',')[int(intra_ins_count - intra_ins_counter)].split(':')[0]
            coord1 = int(out.split('\t')[12].split(' ')[1].split(',')[int(intra_ins_count -
                                                                          intra_ins_counter)].split(':')[1].split('-')[0])
            coord2 = int(out.split('\t')[12].split(' ')[1].split(',')[int(intra_ins_count -
                                                                          intra_ins_counter)].split(':')[1].split('-')[1])
            pair = '.'
            uniqidx = bp_uname + '~' + str(chrom) + ':' + str(coord1) + '-' + str(coord2)
            uniqname = read_name + '~' + bp_uname
            final.append(
                str(read_name) + '\t' + str(bp1) + '\t' + str(bp2) + '\t' + str(bp_name) + ' 99.99' + '~' +
                str(strands) + '\t' + str(chrom) + '\t' + str('-') + '\t' + str(uniqidx) + '\t' + str(pair) + '\t' +
                str(uniqname) + '\t' + str(signature)
            )
            intra_ins_counter = intra_ins_counter - 1
        elif bp_name == 'Intra-Ins(1)':
            chrom = out.split('\t')[12].split(' ')[1].split(',')[int(intra_ins_count - intra_ins_counter)].split(':')[0]
            coord1 = int(out.split('\t')[12].split(' ')[1].split(',')[int(intra_ins_count -
                                                                          intra_ins_counter)].split(':')[1].split('-')[0])
            coord2 = int(out.split('\t')[12].split(' ')[1].split(',')[int(intra_ins_count -
                                                                          intra_ins_counter)].split(':')[1].split('-')[1])
            intrains_size = out.split('\t')[12].split(' ')[0].split(',')[int(intra_ins_count - intra_ins_counter)].split(')')[1]
            pair = 'Intra1'
            uniqidx = bp_uname + '~' + str(chrom) + ':' + str(coord1) + '-' + str(coord2)
            uniqname = read_name + '~' + bp_uname
            if int(intrains_size) >= minlen:
                final.append(
                    str(read_name) + '\t' + str(bp1) + '\t' + str(bp2) + '\t' + str(bp_name) + ' ' +
                    str(intrains_size) + '~' + str(strands) + '\t' + str(chrom) + '\t' + str('-') + '\t' +
                    str(uniqidx) + '\t' + str(pair) + '\t' + str(uniqname) + '\t' + str(signature)
                )
            intra_ins_counter = intra_ins_counter - 1
        elif bp_name == 'Intra-Ins(2)':
            chrom = out.split('\t')[12].split(' ')[1].split(',')[int(intra_ins_count - intra_ins_counter)].split(':')[0]
            coord1 = int(out.split('\t')[12].split(' ')[1].split(',')[int(intra_ins_count -
                                                                          intra_ins_counter)].split(':')[1].split('-')[0])
            coord2 = int(out.split('\t')[12].split(' ')[1].split(',')[int(intra_ins_count -
                                                                          intra_ins_counter)].split(':')[1].split('-')[1])
            intrains_size = out.split('\t')[12].split(' ')[0].split(',')[int(intra_ins_count - intra_ins_counter)].split(')')[1]
            pair = 'Intra2'
            uniqidx = bp_uname + '~' + str(chrom) + ':' + str(coord1) + '-' + str(coord2)
            uniqname = read_name + '~' + bp_uname
            if int(intrains_size) >= minlen:
                final.append(
                    str(read_name) + '\t' + str(bp1) + '\t' + str(bp2) + '\t' + str(bp_name) + ' ' +
                    str(intrains_size) + '~' + str(strands) + '\t' + str(chrom) + '\t' + str('-') + '\t' +
                    str(uniqidx) + '\t' + str(pair) + '\t' + str(uniqname) + '\t' + str(signature)
                )
            intra_ins_counter = intra_ins_counter - 1
        elif bp_name == 'Inter-Ins(1)':
            chrom1 = out.split('\t')[13].split(' ')[1].split(',')[int(inter_ins_count -
                                                                      inter_ins_counter)].split('~')[0].split(':')[0]
            coord1_1 = out.split('\t')[13].split(' ')[1].split(',')[int(inter_ins_count -
                                                                        inter_ins_counter)].split('~')[0].split(':')[1]
            chrom2 = out.split('\t')[13].split(' ')[1].split(',')[int(inter_ins_count -
                                                                      inter_ins_counter)].split('~')[1].split(':')[0]
            coord2_1 = out.split('\t')[13].split(' ')[1].split(',')[int(inter_ins_count -
                                                                        inter_ins_counter)].split('~')[1].split(':')[1]
            interins_size = out.split('\t')[13].split(' ')[0].split(',')[int(inter_ins_count - inter_ins_counter)].split(')')[1]
            pair = 'Inter1'
            uniqidx = bp_uname + '~' + str(chrom1) + ':' + str(coord1_1) + '~' + str(chrom2) + ':' + str(coord2_1)
            uniqname = read_name + '~' + bp_uname
            if int(interins_size) >= minlen:
                final.append(
                    str(read_name) + '\t' + str(bp1) + '\t' + str(bp2) + '\t' + str(bp_name) + ' ' +
                    str(interins_size) + '~' + str(strands) + '\t' + str(chrom1) + '\t' + str('-') + '\t' +
                    str(uniqidx) + '\t' + str(pair) + '\t' + str(uniqname) + '\t' + str(signature)
                )
            inter_ins_counter = inter_ins_counter - 1
        elif bp_name == 'Inter-Ins(2)':
            chrom1 = out.split('\t')[13].split(' ')[1].split(',')[int(inter_ins_count -
                                                                      inter_ins_counter)].split('~')[0].split(':')[0]
            coord1_1 = out.split('\t')[13].split(' ')[1].split(',')[int(inter_ins_count -
                                                                        inter_ins_counter)].split('~')[0].split(':')[1]
            chrom2 = out.split('\t')[13].split(' ')[1].split(',')[int(inter_ins_count -
                                                                      inter_ins_counter)].split('~')[1].split(':')[0]
            coord2_1 = out.split('\t')[13].split(' ')[1].split(',')[int(inter_ins_count -
                                                                        inter_ins_counter)].split('~')[1].split(':')[1]
            interins_size = out.split('\t')[13].split(' ')[0].split(',')[int(inter_ins_count - inter_ins_counter)].split(')')[1]
            pair = 'Inter2'
            uniqidx = bp_uname + '~' + str(chrom1) + ':' + str(coord1_1) + '~' + str(chrom2) + ':' + str(coord2_1)
            uniqname = read_name + '~' + bp_uname
            if int(interins_size) >= minlen:
                final.append(
                    str(read_name) + '\t' + str(bp1) + '\t' + str(bp2) + '\t' + str(bp_name) + ' ' +
                    str(interins_size) + '~' + str(strands) + '\t' + str(chrom1) + '\t' + str('-') + '\t' +
                    str(uniqidx) + '\t' + str(pair) + '\t' + str(uniqname) + '\t' + str(signature)
                )
            inter_ins_counter = inter_ins_counter - 1
        elif bp_name == 'InterTx':
            chrom1 = out.split('\t')[14].split(' ')[1].split(',')[int(inter_tx_count -
                                                                      inter_tx_counter)].split('~')[0].split(':')[0]
            coord1_1 = out.split('\t')[14].split(' ')[1].split(',')[int(inter_tx_count -
                                                                        inter_tx_counter)].split('~')[0].split(':')[1]
            chrom2 = out.split('\t')[14].split(' ')[1].split(',')[int(inter_tx_count -
                                                                      inter_tx_counter)].split('~')[1].split(':')[0]
            coord2_1 = out.split('\t')[14].split(' ')[1].split(',')[int(inter_tx_count -
                                                                        inter_tx_counter)].split('~')[1].split(':')[1]
            pair = '.'
            uniqidx = bp_uname + '~' + str(chrom1) + ':' + str(coord1_1) + '~' + str(chrom2) + ':' + str(coord2_1)
            uniqname = read_name + '~' + bp_uname
            final.append(
                str(read_name) + '\t' + str(bp1) + '\t' + str(bp2) + '\t' + str(bp_name) + ' 99.99' + '~' +
                str(strands) + '\t' + str(chrom1) + '\t' + str('-') + '\t' + str(uniqidx) + '\t' + str(pair) + '\t' +
                str(uniqname) + '\t' + str(signature)
            )
            inter_tx_counter = inter_tx_counter - 1
        else:
            raise Exception("Error: Unidentifiable SV")
    return final


# Generate SV signature
def getsignature(bp1, name, querymap, sign, realnmaps, elist, bitlist, bps_no, complex_sv, nchrom, short, piden, mismatch,
                 gap_ratio, strandness, sig_index):
    signlist = sign.split(',')
    if name == 'S-Nov_Ins':
        signature = ','.join(signlist[0:3]) + ',0,0,' + str(elist[0]) + ',0,' + str(bitlist[0]) + ',0,' + \
                    str(complex_sv) + ',' + str(realnmaps) + ',' + str(bps_no) + ',' + str(nchrom) + ',' + str(short) + ',' + \
                    str(piden[0]) + ',0,' + str(mismatch[0]) + ',0,' + str(gap_ratio[0]) + ',0'
        strands = str(strandness[0])
    elif name == 'E-Nov_Ins':
        signature = ','.join(signlist[-3:]) + ',0,0,' + str(elist[-1]) + ',0,' + str(bitlist[-1]) + ',0,' + str(complex_sv) + \
                    ',' + str(realnmaps) + ',' + str(bps_no) + ',' + str(nchrom) + ',' + str(short) + ',' + str(piden[-1]) + \
                    ',0,' + str(mismatch[-1]) + ',0,' + str(gap_ratio[-1]) + ',0'
        strands = str(strandness[-1])
    else:
        maps = ast.literal_eval(querymap)
        nmaps = len(maps)
        signature = ''
        strands = ''
        for i in range(nmaps):
            if int(maps[i][1]) == bp1:
                signature = ','.join(signlist[sig_index[i]:sig_index[i] + 5]) + ',' + str(elist[i]) + ',' + str(elist[i + 1]) + \
                            ',' + str(bitlist[i]) + ',' + str(bitlist[i + 1]) + ',' + str(complex_sv) + ',' + str(realnmaps) + \
                            ',' + str(bps_no) + ',' + str(nchrom) + ',' + str(short) + ',' + str(piden[i]) + ',' + \
                            str(piden[i + 1]) + ',' + str(mismatch[i]) + ',' + str(mismatch[i + 1]) + ',' + str(gap_ratio[i]) + \
                            ',' + str(gap_ratio[i + 1])
                strands = str(strandness[i]) + ',' + str(strandness[i + 1])
                break
    return signature, strands



"""
SV detection algorithm.

Copyright (C) 2021 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""


def sv_detect(subdata, splitpct, minalign, gapdict):
    c = len(subdata)  # Number of entries
    if c == 0:
        out1, out2 = '', ''
    else:
        if c == 1:
            sv_status = single_alignment_check(subdata, splitpct, minalign)
        else:
            subdata = reduce_multi_map(subdata)
            c = len(subdata)  # Recalculate number of entries
            sv_status = True
        if sv_status:
            tmpread = subdata
            actual_no_maps = subdata[0].split('\t')[14].split('=')[1]
            chromocollect = tmpread[0].split('\t')[15].split('c')[0]
            # Set minimum length for read terminal breakends
            termlength = 100
            # Temporary set SV len cutoff
            ncutoff = 19
            j = range(c)
            p = range(c - 1)
            # Define SV flags
            strand_total, evalue_total, bitscore_ratio_total, piden_total, mismatch_ratio_total, gap_ratio_total, query, \
                subject, chromorder, qurygaps, complex_sv, iso_nov_ins, iso_ins_size, iso_ins_range, sv_range, intra_ins, \
                intra_ins_range, dele, del_size, del_range, tdup, tdup_range, nov_ins, ins_size, ins_range, inv, inv_range, \
                complex_nov_ins, complex_ins_size, inter_ins, inter_ins_range, intertx, intertx_break = ([] for i in range(33))
            # Satellite elements counter, Short alignments counter, Ref chromosome counter, Strandness counter, Second bp markers
            a = f = q = s = g = gtx = 0
            # Choose tentative reference chromosome and strandness
            refchr = tmpread[0].split('\t')[0]
            refstrand = tmpread[0].split('\t')[7]
            # Retrieving read length
            readlength = int(tmpread[0].split('\t')[8])
            # Define short length alignment (Arbitrary definition as <5% of read length)
            t = int(readlength) * 0.05
            for i in j:  # Collecting alignment information and filtering
                # Scale evalue 100000*, round to 5 dec places
                evalue = round(float(tmpread[i].split('\t')[9]) * 100000, 5)
                # Calculate bitscore ratio
                bitscore_ratio = round(float(tmpread[i].split('\t')[10]) / int(tmpread[i].split('\t')[6]), 5)
                # Collect alignment percentage identity
                piden = tmpread[i].split('\t')[11]
                # Calculate mismatch ratio
                mismatch_ratio = round(float(tmpread[i].split('\t')[12]) / int(tmpread[i].split('\t')[6]), 5)
                # Calculate gap ratio
                gap_ratio = round(float(tmpread[i].split('\t')[13]) / int(tmpread[i].split('\t')[6]), 5)
                # Get strandness
                strand = tmpread[i].split('\t')[7]
                # Collect all alignment info
                evalue_total.append(str(evalue))
                bitscore_ratio_total.append(str(bitscore_ratio))
                piden_total.append(str(piden))
                mismatch_ratio_total.append(str(mismatch_ratio))
                gap_ratio_total.append(str(gap_ratio))
                strand_total.append(str(strand))
                # Collecting query and subject information
                query.append([int(tmpread[i].split('\t')[5]), int(tmpread[i].split('\t')[5]) + int(tmpread[i].split('\t')[6])])
                if tmpread[i].split('\t')[7] == '+':
                    subject.append([int(tmpread[i].split('\t')[1]), int(tmpread[i].split('\t')[1]) +
                                    int(tmpread[i].split('\t')[2])])
                elif tmpread[i].split('\t')[7] == '-':
                    subject.append([int(tmpread[i].split('\t')[1]) + int(tmpread[i].split('\t')[2]),
                                    int(tmpread[i].split('\t')[1])])
                # Filtering out strict telomeric and centromere regions
                chrm = tmpread[i].split('\t')[0].strip()
                chromorder.append(chrm)
                subjstart = int(tmpread[i].split('\t')[1])
                subjend = int(tmpread[i].split('\t')[1]) + int(tmpread[i].split('\t')[2])
                if gapdict is not None:
                    try:
                        for v in gapdict[chrm]:
                            if v[0] <= subjstart <= v[1] or v[0] <= subjend <= v[1]:
                                a += 1
                    except KeyError:
                        pass
                # Count short alignments (Arbitrary definition as <5% of read length)
                if int(tmpread[i].split('\t')[6]) < t:
                    f += 1
                # Count support for ref chromosome and strandness
                if tmpread[i].split('\t')[0] == refchr:
                    q += 1
                if tmpread[i].split('\t')[7] == refstrand:
                    s += 1
            # Calculating query middle gaps length
            for i in p:
                qurygaps.append(query[i + 1][0] - query[i][1])
            # Check for Satellite element
            if a == c:  # All alignments lies within satellite region
                sv_status = False
            elif 0 < a:
                sv_status = False
            # Check for short alignment lengths
            short = float(f) / c
            # Defining major strandness
            majstrand = ''  # Dummy
            if float(s) / c >= 0.5:
                majstrand = refstrand
            else:
                if refstrand == '+':
                    majstrand = '-'
                elif refstrand == '-':
                    majstrand = '+'
            # Counting number of inversions
            if majstrand == refstrand:
                if (c - s) > 1:
                    complex_sv.append('Multi_inv')
            else:
                if s > 1:
                    complex_sv.append('Multi_inv')
            # Calculating start/end gaps length and evaluation based on 100bp or more than ncutoff bases
            startgap = int(tmpread[0].split('\t')[5])
            endgap = int(tmpread[int(c) - 1].split('\t')[8]) - int(tmpread[int(c) - 1].split('\t')[6]) - int(
                tmpread[int(c) - 1].split('\t')[5])
            if startgap > termlength and startgap > ncutoff:
                iso_nov_ins.append('S-Nov-Ins')
                iso_ins_size.append('S-Nov-Ins' + str(startgap))
                iso_ins_range.append(str(tmpread[0].split('\t')[0].strip()) + ':' + str(subject[0][0]))
                sv_range.append('S-Nov_Ins:1-' + str(query[0][0] + 1))
            if endgap > termlength and endgap > ncutoff:
                iso_nov_ins.append('E-Nov-Ins')
                iso_ins_size.append('E-Nov-Ins' + str(endgap))
                iso_ins_range.append(str(tmpread[-1].split('\t')[0].strip()) + ':' + str(subject[-1][1]))
                sv_range.append('E-Nov_Ins:' + str(query[-1][1] - 1) + '-' + str(readlength))
            # Begin scanning all alignments on read
            if sv_status:
                for i in p:
                    sbp = 0
                    qurygap = query[i + 1][0] - query[i][1]
                    # Same chromosome
                    if tmpread[i].split('\t')[0] == tmpread[i + 1].split('\t')[0]:
                        # Same strandness (1)
                        if tmpread[i].split('\t')[7] == '+' and tmpread[i + 1].split('\t')[7] == '+':
                            if i != int(g - 1):  # Prevent duplicate record
                                # Gapstudy
                                subjgap = subject[i + 1][0] - subject[i][1]
                                if subjgap > ncutoff:  # If gap is positive and more than ncutoff
                                    # Subject gap needs to be at least twice the size of query gap
                                    if float(subjgap) / (qurygap + 1) >= 2:
                                        # First breakpoint detected
                                        # Detecting second breakpoint
                                        for u in range(i + 2, c):
                                            size = query[u][0] - query[i][1]
                                            if tmpread[i].split('\t')[0] == tmpread[u].split('\t')[0] and \
                                                    tmpread[i].split('\t')[7] == tmpread[u].split('\t')[7] and \
                                                    any(subject[i][1] < y < min(subject[i+1][0], subject[i][1]+size+1000) for y
                                                        in subject[u]):
                                                # Second breakpoint detected, insertion event
                                                sbp = 1
                                                g = int(u)
                                                break
                                        if sbp == 1:
                                            # Second breakpoint present
                                            intra_ins.append('Intra-Ins(2)' + str(query[g][0] - query[i][1]))
                                            # Duplicate record for Intra-Ins(1)
                                            intra_ins.append('Intra-Ins(2)' + str(query[g][0] - query[i][1]))
                                            intra_ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' +
                                                                   str(subject[i][1]) + '-' + str(subject[i + 1][0]))
                                            intra_ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' +
                                                                   str(subject[g - 1][1]) + '-' + str(subject[g][0]))
                                            sv_range.append('Intra-Ins(1):' + str(query[i][1]) + '-' + str(query[i + 1][0]) +
                                                            ',' + 'Intra-Ins(2):' + str(query[g - 1][1]) + '-' +
                                                            str(query[g][0]))
                                        elif sbp == 0:
                                            if subjgap - qurygap <= 100000:
                                                dele.append('Del')
                                                del_size.append('Del' + str(subjgap-qurygap))
                                                del_range.append(
                                                    str(tmpread[i].split('\t')[0].strip()) + ':' + str(subject[i][1]) + '-' + str(
                                                        subject[i + 1][0]))
                                                sv_range.append('Del:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                                            else:
                                                intra_ins.append('Intra-Ins')
                                                intra_ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' +
                                                                       str(subject[i][1]) + '-' + str(subject[i + 1][0]))
                                                sv_range.append('Intra-Ins:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                                # If gap is negative and smaller than 100000 (Arbitrary tandemDup size gauge)
                                elif -100000 < subjgap < -20:
                                    tdup.append('TDupl')
                                    tdup_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' + str(subject[i][1]) + '-' +
                                                      str(subject[i + 1][0]))
                                    sv_range.append('TDupl:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                                # If gap is negative and larger than or equal to 100000
                                elif subjgap <= -100000:
                                    # Detecting second breakpoint
                                    for u in range(i + 2, c):
                                        size = query[u][0] - query[i][1]
                                        if tmpread[i].split('\t')[0] == tmpread[u].split('\t')[0] and \
                                                tmpread[i].split('\t')[7] == tmpread[u].split('\t')[7] and \
                                                any(subject[i][1] < y < subject[i][1]+size+1000 for y in subject[u]):  # May
                                            # still contain large deletion
                                            sbp = 1
                                            g = int(u)
                                            break
                                    if sbp == 1:
                                        # Second breakpoint present
                                        intra_ins.append('Intra-Ins(2)' + str(query[g][0] - query[i][1]))
                                        # Duplicate record for Intra-Ins(1)
                                        intra_ins.append('Intra-Ins(2)' + str(query[g][0] - query[i][1]))
                                        intra_ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' +
                                                               str(subject[i][1]) + '-' + str(subject[i + 1][0]))
                                        intra_ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' +
                                                               str(subject[g - 1][1]) + '-' + str(subject[g][0]))
                                        sv_range.append('Intra-Ins(1):' + str(query[i][1]) + '-' + str(query[i + 1][0]) + ',' +
                                                        'Intra-Ins(2):' + str(query[g - 1][1]) + '-' + str(query[g][0]))
                                    elif sbp == 0:
                                        intra_ins.append('Intra-Ins')
                                        intra_ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' +
                                                               str(subject[i][1]) + '-' + str(subject[i + 1][0]))
                                        sv_range.append('Intra-Ins:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                                # If query gap more than cutoff and subject gap is positive
                                if qurygap > ncutoff and subjgap >= 0:
                                    if float(qurygap) / (subjgap + 1) >= 2:  # Test for novel insertion ratio
                                        nov_ins.append('Nov-Ins')
                                        ins_size.append('Nov-Ins' + str(qurygap-subjgap))
                                        ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' + str(subject[i][1]) +
                                                         '-' + str(int(subject[i][1]) + 1))
                                        sv_range.append('Nov_Ins:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                                # If query gap more than cutoff and subject gap is negative
                                elif qurygap > ncutoff and subjgap < 0:
                                    nov_ins.append('Nov-Ins')
                                    ins_size.append('Nov-Ins' + str(qurygap))
                                    ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' + str(subject[i][1]) + '-' +
                                                     str(int(subject[i][1]) + 1))
                                    sv_range.append('Nov_Ins:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                        # Same strandness (2)
                        elif tmpread[i].split('\t')[7] == '-' and tmpread[i + 1].split('\t')[7] == '-':
                            if i != int(g - 1):  # Prevent duplicate record
                                # Gapstudy
                                subjgap = subject[i][1] - subject[i + 1][0]
                                if subjgap > ncutoff:  # Is gap positive value and more than ncutoff?
                                    if float(subjgap) / (qurygap + 1) >= 2:  # Test for significant gap using ratio
                                        # First breakpoint detected
                                        # Detecting second breakpoint
                                        for u in range(i + 2, c):
                                            size = query[u][0] - query[i][1]
                                            if tmpread[i].split('\t')[0] == tmpread[u].split('\t')[0] and \
                                                    tmpread[i].split('\t')[7] == tmpread[u].split('\t')[7] and \
                                                    any(max(subject[i + 1][0], subject[i][1]-size-1000) < y < subject[i][1] for
                                                        y in subject[u]):
                                                # Second breakpoint detected, insertion event
                                                sbp = 1
                                                g = int(u)
                                                break
                                        if sbp == 1:
                                            # Second breakpoint present
                                            intra_ins.append('Intra-Ins(2)' + str(query[g][0] - query[i][1]))
                                            # Duplicate record for Intra-Ins(1)
                                            intra_ins.append('Intra-Ins(2)' + str(query[g][0] - query[i][1]))
                                            intra_ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' +
                                                                   str(subject[i][1]) + '-' + str(subject[i + 1][0]))
                                            intra_ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' +
                                                                   str(subject[g - 1][1]) + '-' + str(subject[g][0]))
                                            sv_range.append('Intra-Ins(1):' + str(query[i][1]) + '-' + str(query[i + 1][0]) +
                                                            ',' + 'Intra-Ins(2):' + str(query[g - 1][1]) + '-' + str(query[g][0]))
                                        elif sbp == 0:
                                            if subjgap - qurygap <= 100000:
                                                dele.append('Del')
                                                del_size.append('Del' + str(subjgap-qurygap))
                                                del_range.append(
                                                    str(tmpread[i].split('\t')[0].strip()) + ':' + str(subject[i][1]) + '-' + str(
                                                        subject[i + 1][0]))
                                                sv_range.append('Del:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                                            else:
                                                intra_ins.append('Intra-Ins')
                                                intra_ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' +
                                                                       str(subject[i][1]) + '-' + str(subject[i + 1][0]))
                                                sv_range.append('Intra-Ins:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                                # If gap is negative and smaller than 100000 (Arbitrary tandemDup size gauge)
                                elif -100000 < subjgap < -20:
                                    tdup.append('TDupl')
                                    tdup_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' + str(subject[i][1]) + '-' +
                                                      str(subject[i + 1][0]))
                                    sv_range.append('TDupl:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                                # If gap is negative and larger than or equal to 100000
                                elif subjgap <= -100000:
                                    # Detecting second breakpoint
                                    for u in range(i + 2, c):
                                        size = query[u][0] - query[i][1]
                                        if tmpread[i].split('\t')[0] == tmpread[u].split('\t')[0] and \
                                                tmpread[i].split('\t')[7] == tmpread[u].split('\t')[7] and \
                                                any(subject[i][1]-size-1000 < y < subject[i][1] for y in subject[u]):
                                            # May still contain large deletion
                                            sbp = 1
                                            g = int(u)
                                            break
                                    if sbp == 1:
                                        # Second breakpoint present
                                        intra_ins.append('Intra-Ins(2)' + str(query[g][0] - query[i][1]))
                                        # Duplicate record for Intra-Ins(1)
                                        intra_ins.append('Intra-Ins(2)' + str(query[g][0] - query[i][1]))
                                        intra_ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' +
                                                               str(subject[i][1]) + '-' + str(subject[i + 1][0]))
                                        intra_ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' +
                                                               str(subject[g - 1][1]) + '-' + str(subject[g][0]))
                                        sv_range.append('Intra-Ins(1):' + str(query[i][1]) + '-' + str(query[i + 1][0]) + ',' +
                                                        'Intra-Ins(2):' + str(query[g - 1][1]) + '-' + str(query[g][0]))
                                    elif sbp == 0:
                                        intra_ins.append('Intra-Ins')
                                        intra_ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' +
                                                               str(subject[i][1]) + '-' + str(subject[i + 1][0]))
                                        sv_range.append('Intra-Ins:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                                # If query gap more than cutoff and subject gap is positive
                                if qurygap > ncutoff and subjgap >= 0:
                                    if float(qurygap) / (subjgap + 1) >= 2:  # Test for novel insertion ratio
                                        nov_ins.append('Nov-Ins')
                                        ins_size.append('Nov-Ins' + str(qurygap-subjgap))
                                        ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' + str(subject[i][1]) + '-' +
                                                         str(int(subject[i][1]) + 1))
                                        sv_range.append('Nov_Ins:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                                # If query gap more than cutoff and subject gap is negative
                                elif qurygap > ncutoff and subjgap < 0:
                                    nov_ins.append('Nov-Ins')
                                    ins_size.append('Nov-Ins' + str(qurygap))
                                    ins_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' + str(subject[i][1]) + '-' +
                                                     str(int(subject[i][1]) + 1))
                                    sv_range.append('Nov_Ins:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                        # Different strandness (1)
                        elif tmpread[i].split('\t')[7] == '+' and tmpread[i + 1].split('\t')[7] == '-':  # Inversion
                            if i != int(g - 1):
                                for u in range(i + 2, c):
                                    size = query[u][0] - query[i][1]
                                    if tmpread[i].split('\t')[0] == tmpread[u].split('\t')[0] and \
                                            tmpread[i].split('\t')[7] == tmpread[u].split('\t')[7] and \
                                            any(subject[i][1] < y < subject[i][1]+size+1000 for y in subject[u]):
                                        # Second breakpoint detected, paired inversion event
                                        sbp = 1
                                        g = int(u)
                                        break
                                if sbp == 1:
                                    # Second breakpoint present
                                    inv.append('Pair_Inv' + str(query[g][0] - query[i][1]))
                                    # Duplicate record for Inv(1)
                                    inv.append('Pair_Inv' + str(query[g][0] - query[i][1]))
                                    inv_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' + str(subject[i][1]) + '-' +
                                                     str(subject[i + 1][0]))
                                    inv_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' + str(subject[g - 1][1]) + '-' +
                                                     str(subject[g][0]))
                                    sv_range.append('Inv(1):' + str(query[i][1]) + '-' + str(query[i + 1][0]) + ',' + 'Inv(2):' +
                                                    str(query[g - 1][1]) + '-' + str(query[g][0]))
                                elif sbp == 0:
                                    inv.append('Unpair_Inv')
                                    inv_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' + str(subject[i][1]) + '-' +
                                                     str(subject[i + 1][0]))
                                    sv_range.append('Inv:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                            if qurygap > ncutoff:  # If query gap more than ncutoff
                                complex_nov_ins.append('Inv-Nov-Ins')
                                complex_sv.append('Inv-Nov_Ins')
                                complex_ins_size.append('Inv-Nov_Ins' + str(qurygap))
                        # Different Strandness(2)
                        elif tmpread[i].split('\t')[7] == '-' and tmpread[i + 1].split('\t')[7] == '+':  # Inversion
                            if i != int(g - 1):
                                for u in range(i + 2, c):
                                    size = query[u][0] - query[i][1]
                                    if tmpread[i].split('\t')[0] == tmpread[u].split('\t')[0] and \
                                            tmpread[i].split('\t')[7] == tmpread[u].split('\t')[7] and \
                                            any(subject[i][1]-size-1000 < y < subject[i][1] for y in subject[u]):
                                        # Second breakpoint detected, paired inversion event
                                        sbp = 1
                                        g = int(u)
                                        break
                                if sbp == 1:
                                    # Second breakpoint present
                                    inv.append('Pair_Inv' + str(query[g][0] - query[i][1]))
                                    # Duplicate record for Inv(1)
                                    inv.append('Pair_Inv' + str(query[g][0] - query[i][1]))
                                    inv_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' + str(subject[i][1]) + '-' +
                                                     str(subject[i + 1][0]))
                                    inv_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' + str(subject[g - 1][1]) + '-' +
                                                     str(subject[g][0]))
                                    sv_range.append('Inv(1):' + str(query[i][1]) + '-' + str(query[i + 1][0]) + ',' + 'Inv(2):' +
                                                    str(query[g - 1][1]) + '-' + str(query[g][0]))
                                elif sbp == 0:
                                    inv.append('Unpair_Inv')
                                    inv_range.append(str(tmpread[i].split('\t')[0].strip()) + ':' + str(subject[i][1]) + '-' +
                                                     str(subject[i + 1][0]))
                                    sv_range.append('Inv:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                            if qurygap > ncutoff:  # If query gap more than ncutoff
                                complex_nov_ins.append('Inv-Nov-Ins')
                                complex_sv.append('Inv-Nov_Ins')
                                complex_ins_size.append('Inv-Nov_Ins' + str(qurygap))
                    # Different chromosome
                    elif tmpread[i].split('\t')[0] != tmpread[i + 1].split('\t')[0]:
                        # Strandness+
                        if tmpread[i].split('\t')[7] == '+':
                            if i != int(gtx - 1):
                                for u in range(i + 2, c):
                                    size = query[u][0] - query[i][1]
                                    if tmpread[i].split('\t')[0] == tmpread[u].split('\t')[0] and \
                                            tmpread[i].split('\t')[7] == tmpread[u].split('\t')[7] and \
                                            any(subject[i][1] < y < subject[i][1]+size+1000 for y in subject[u]):  # Detecting
                                        # second breakpoint
                                        # Second breakpoint detected, insertion event
                                        sbp = 1
                                        gtx = int(u)
                                        break
                                if sbp == 1:
                                    # Second breakpoint present
                                    inter_ins.append('Inter-Ins(2)' + str(query[gtx][0] - query[i][1]))
                                    # Duplicate record for Inter-Ins(1)
                                    inter_ins.append('Inter-Ins(2)' + str(query[gtx][0] - query[i][1]))
                                    inter_ins_range.append(str(tmpread[i].split('\t')[0]) + ':' + str(subject[i][1]) + '~' +
                                                           str(tmpread[i + 1].split('\t')[0]) + ':' + str(subject[i + 1][0]))
                                    inter_ins_range.append(str(tmpread[gtx - 1].split('\t')[0]) + ':' +
                                                           str(subject[gtx - 1][1]) + '~' + str(tmpread[gtx].split('\t')[0]) +
                                                           ':' + str(subject[gtx][0]))
                                    sv_range.append('Inter-Ins(1):' + str(query[i][1]) + '-' + str(query[i + 1][0]) + ',' +
                                                    'Inter-Ins(2):' + str(query[gtx - 1][1]) + '-' + str(query[gtx][0]))
                                elif sbp == 0:
                                    intertx.append('InterTx')
                                    intertx_break.append(str(tmpread[i].split('\t')[0]) + ':' + str(subject[i][1]) + '~' +
                                                         str(tmpread[i + 1].split('\t')[0]) + ':' + str(subject[i + 1][0]))
                                    sv_range.append('InterTx:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                                if qurygap > ncutoff:  # If query gap more than ncutoff
                                    complex_nov_ins.append('Inter-Nov-Ins')
                                    complex_sv.append('Inter-Nov_Ins')
                                    complex_ins_size.append('Inter-Nov_Ins' + str(qurygap))
                        # Strandness-
                        if tmpread[i].split('\t')[7] == '-':
                            if i != int(gtx - 1):
                                for u in range(i + 2, c):
                                    size = query[u][0] - query[i][1]
                                    if tmpread[i].split('\t')[0] == tmpread[u].split('\t')[0] and \
                                            tmpread[i].split('\t')[7] == tmpread[u].split('\t')[7] and \
                                            any(subject[i][1]-size-1000 < y < subject[i][1] for y in subject[u]):  # Detecting
                                        # second breakpoint
                                        # Second breakpoint detected, insertion event
                                        sbp = 1
                                        gtx = int(u)
                                        break
                                if sbp == 1:
                                    # Second breakpoint present
                                    inter_ins.append('Inter-Ins(2)' + str(query[gtx][0] - query[i][1]))
                                    # Duplicate record for Inter-Ins(1)
                                    inter_ins.append('Inter-Ins(2)' + str(query[gtx][0] - query[i][1]))
                                    inter_ins_range.append(str(tmpread[i].split('\t')[0]) + ':' + str(subject[i][1]) + '~' +
                                                           str(tmpread[i + 1].split('\t')[0]) + ':' + str(subject[i + 1][0]))
                                    inter_ins_range.append(str(tmpread[gtx - 1].split('\t')[0]) + ':' +
                                                           str(subject[gtx - 1][1]) + '~' + str(tmpread[gtx].split('\t')[0]) +
                                                           ':' + str(subject[gtx][0]))
                                    sv_range.append('Inter-Ins(1):' + str(query[i][1]) + '-' + str(query[i + 1][0]) + ',' +
                                                    'Inter-Ins(2):' + str(query[gtx - 1][1]) + '-' + str(query[gtx][0]))
                                elif sbp == 0:
                                    intertx.append('InterTx')
                                    intertx_break.append(str(tmpread[i].split('\t')[0]) + ':' + str(subject[i][1]) + '~' + str(
                                        tmpread[i + 1].split('\t')[0]) + ':' + str(subject[i + 1][0]))
                                    sv_range.append('InterTx:' + str(query[i][1]) + '-' + str(query[i + 1][0]))
                                if qurygap > ncutoff:  # If query gap more than ncutoff
                                    complex_nov_ins.append('Inter-Nov-Ins')
                                    complex_sv.append('Inter-Nov_Ins')
                                    complex_ins_size.append('Inter-Nov_Ins' + str(qurygap))
            if dele == [] and iso_nov_ins == [] and nov_ins == [] and complex_nov_ins == [] and tdup == [] and inv == [] and \
                    intra_ins == [] and inter_ins == [] and intertx == [] and complex_sv == []:
                sv_status = False
            total_sv_complex = ' '.join(dele) + ' ' + ' '.join(iso_nov_ins) + ' ' + ' '.join(nov_ins) + ' ' + ' '.join(
                tdup) + ' ' + ' '.join(inv) + ' ' + ' '.join(intra_ins) + ' ' + ' '.join(inter_ins) + ' ' + ' '.join(intertx)
            if sv_status:
                querypercent = []
                qurygappercent = []
                fullquerypercent = []
                for i in p:  # Calculating gap percentages
                    qurygappercent.append(str((float(qurygaps[i]) / readlength) * 100))
                startgappercent = str((float(startgap) / readlength) * 100)
                qurygappercent.append(str((float(endgap) / readlength) * 100))
                for i in j:
                    querypercent.append(
                        str((float(tmpread[i].split('\t')[6]) / readlength) * 100))  # Calculating query map percentage
                fullquerypercent.append('%.1f' % float(startgappercent) + '%')
                for i in j:
                    fullquerypercent.append('(' + '%.1f' % float(querypercent[i]) + '%)')
                    fullquerypercent.append('%.1f' % float(qurygappercent[i]) + '%')
                out1 = ''
                out2 = str(tmpread[0].split('\t')[4].strip()) + '\t' + str(chromocollect) + '\t' + \
                    str(actual_no_maps) + ' maps' + '\t' + ','.join(fullquerypercent) + '\t' + \
                    ','.join(evalue_total) + '\t' + ','.join(bitscore_ratio_total) + '\t' + ','.join(del_size) + ' ' + \
                    ','.join(del_range) + '\t' + ','.join(iso_ins_size) + ' ' + ','.join(iso_ins_range) + '\t' + \
                    ','.join(ins_size) + ' ' + ','.join(ins_range) + '\t' + ','.join(complex_ins_size) + '\t' + \
                    ','.join(tdup) + ' ' + ','.join(tdup_range) + '\t' + ','.join(inv) + ' ' + ','.join(inv_range) + '\t' + \
                    ','.join(intra_ins) + ' ' + ','.join(intra_ins_range) + '\t' + ','.join(inter_ins) + ' ' + \
                    ','.join(inter_ins_range) + '\t' + ','.join(intertx) + ' ' + ','.join(intertx_break) + '\t' + \
                    ','.join(strand_total) + '\t' + str(short) + '\t' + ','.join(complex_sv) + '\t' + \
                    ', '.join(total_sv_complex.split()) + '\t' + ','.join(sv_range) + '\t' + str(query) + '\t' + \
                    ','.join(piden_total) + '\t' + ','.join(mismatch_ratio_total) + '\t' + ','.join(gap_ratio_total)
            else:
                out1, out2 = '', ''
        else:
            out1, out2 = '', ''
    return out1, out2


# Out1 columns: 1=readname, 2=readlength, 3=total_chromosome(s), 4=No._of_maps, 5=query_signature, 6=evalue, 7=bitscore,
# 8=deletion, 9=Isolated_insertion, 10=Insertion, 11=complex_insertion, 12=tandemDupl, 13=inversion, 14=intra_insertion,
# 15=inter_insertion, 16=inter_tx, 17=strandness, 18=short_frag_perc, 19=satellite_perc, 20=complex_SV, 21=Total_SV_complex,
# 22=sv_range, 23=chromosome_order, 24=subject_string, 25=query_string, 26=%identity, 27=mismatch, 28=gap_ratio

# out1 = str(tmpread[0].split('\t')[4].strip()) + '\t' + str(readlength) + '\t' + str(chromocollect) + '\t' + \
#     str(actual_no_maps) + ' maps' + '\t' + ','.join(fullquerypercent) + '\t' + ','.join(evalue_total) + '\t' + \
#     ','.join(bitscore_ratio_total) + '\t' + ','.join(del_size) + ' ' + ','.join(del_range) + '\t' + \
#     ','.join(iso_ins_size) + ' ' + ','.join(iso_ins_range) + '\t' + ','.join(ins_size) + ' ' + \
#     ','.join(ins_range) + '\t' + ','.join(complex_ins_size) + '\t' + ','.join(tdup) + ' ' + \
#     ','.join(tdup_range) + '\t' + ','.join(inv) + ' ' + ','.join(inv_range) + '\t' + ','.join(intra_ins) + ' ' + \
#     ','.join(intra_ins_range) + '\t' + ','.join(inter_ins) + ' ' + ','.join(inter_ins_range) + '\t' + \
#     ','.join(intertx) + ' ' + ','.join(intertx_break) + '\t' + ','.join(strand_total) + '\t' + \
#     str(short) + '\t' + sat + '\t' + ','.join(complex_sv) + '\t' + ", ".join(total_sv_complex.split()) + '\t' + \
#     ','.join(sv_range) + '\t' + str(chromorder) + '\t' + ''.join(str(subject)) + '\t' + str(query) + '\t' + \
#     ','.join(piden_total) + '\t' + ','.join(mismatch_ratio_total) + '\t' + ','.join(gap_ratio_total)

# Out2: 0=readname, 1=total_chromosome(s), 2=No._of_maps, 3=query_signature, 4=evalue, 5=bitscore, 6=deletion,
# 7=Isolated_insertion, 8=Insertion, 9=complex_insertion, 10=tandemDupl, 11=inversion, 12=intra_insertion, 13=inter_insertion,
# 14=inter_tx, 15=strandness, 16=short_frag_perc, 17=complex_SV, 18=Total_SV_complex, 19=sv_range, 20=query_string, 21=%identity,
# 22=mismatch, 23=gap_ratio


# Check if single alignment covers (1-splitpct) of read length
def single_alignment_check(subdata, splitpct, minalign):
    align_percent = float(subdata[0].split('\t')[6]) / float(subdata[0].split('\t')[8])
    align_len = int(subdata[0].split('\t')[6])
    if align_percent >= (float(1)-splitpct) or align_len < minalign:
        return False
    else:
        return True


# Reduce mapping of multi-mapping reads
def reduce_multi_map(subdata):
    max_seg = 20
    if len(subdata) > max_seg:
        sortdictlen = {}
        for line in subdata:
            sortdictlen[line] = int(line.split('\t')[6])
        # Sorting according to alignment length
        subdata2 = [key for (key, value) in sorted(sortdictlen.items(), key=lambda x: x[1], reverse=True)]
        subdata3 = subdata2[0:max_seg]  # Selecting top x longest alignments
        # Sort back according to query start
        sortdict = {}
        for e in subdata3:
            sortdict[int(e.split('\t')[5])] = e
        subdata4 = [value for (key, value) in sorted(sortdict.items())]
        return subdata4
    else:
        return subdata


# For future implementation to limit multi-mapping
def discard_multi_map(actual_no_maps, multi_limit):
    # Throw away multi-mapper reads
    if int(actual_no_maps) >= multi_limit:
        return False
    else:
        return True



"""
Duplication and transposable element analyzer. 

Copyright (C) 2023 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""


import os
import pysam
import logging
from pybedtools import BedTool
# from subprocess import Popen, PIPE, STDOUT
from nanovar.nv_align import align_mm

def dup_te_analyzer(wk_dir, out_nn, total_out1, score_threshold, ref_dir, ref_path, mm, threads, data_type, st, debug):
    index2reads = get_index2reads(out_nn, score_threshold)
    region_dict, ins_coord_dict, dups = get_region_dict(total_out1)
    read2index = get_ins_sequence(index2reads, region_dict, wk_dir)
    te_ref_file = os.path.join(ref_dir, 'hg38_L1_Alu.fa')
    reads_file = os.path.join(wk_dir, 'ins_seq.fa')
    _, bam_te = align_mm(te_ref_file, reads_file, wk_dir, 'ins_seq', 'te', threads, mm, data_type, st)
    index2te = parse_bam_te(bam_te, wk_dir, read2index)
    unmap_reads_file = os.path.join(wk_dir, 'ins_seq_te-unmap.fa')
    _, bam_ref = align_mm(ref_path, unmap_reads_file, wk_dir, 'unmap', 'ref', threads, mm, data_type, st)
    index2dup, dup_coord = parse_bam_ref(bam_ref, ins_coord_dict, read2index)
    index2dup = update_dups(dups, index2dup, index2reads)
    dup_indexes, _ = filter_dup(index2dup, index2reads, threshold=0.75)
    new_out_nn = update_out_nn(dup_indexes, out_nn, dup_coord)
    if not debug:
        # os.remove(reads_file)
        os.remove(unmap_reads_file)
        os.remove(bam_te)
        os.remove(bam_ref)
        
    return index2te, new_out_nn

def get_index2reads(out_nn, score_threshold):
    # Convert score threshold to nn probability
    nn_score = 1-10**(score_threshold/-10)
    index2reads = {}
    for i in out_nn:
        if i.split('\t')[3].split(' ')[0] in ['Nov_Ins', 'S-Nov_Ins_bp', 'E-Nov_Ins_bp']:
            if float(i.split('\t')[13]) >= nn_score:
                ind = i.split('\t')[6].split('~')[0]
                reads = i.split('\t')[8] + ',' + i.split('\t')[11]
                index2reads[ind] = reads
    return index2reads

def get_region_dict(total_out):
    region_dict = {}
    ins_coord_dict = {}
    dups = {}
    for i in total_out:
        sv_type = i.split('\t')[3]
        if sv_type[:3] in ('S-N', 'E-N', 'Nov'):
            readindex = i.split('\t')[8]
            query_ins_start = int(i.split('\t')[1])
            ins_size = float(i.split('\t')[3].split(' ')[1].split('~')[0])
            chrm = i.split('\t')[6].split('~')[1].split(':')[0]
            coord = int(i.split('\t')[6].split('~')[1].split(':')[1].split('-')[0])
            ins_coord_dict[readindex] = [chrm, coord-ins_size, coord+ins_size]
            if sv_type[:3] == 'S-N':
                region_dict[readindex] = (1, ins_size)
            elif sv_type[:3] == 'E-N':
                region_dict[readindex] = (query_ins_start, query_ins_start + ins_size)
            elif sv_type[:3] == 'Nov':
                region_dict[readindex] = (query_ins_start, query_ins_start + ins_size)
        elif sv_type[:5] == 'TDupl':
            readindex = i.split('\t')[8]
            dups[readindex] = ''
    return region_dict, ins_coord_dict, dups

def get_ins_sequence(index2reads, region_dict, wk_dir):
    bed_str = ''
    read2index = {}
    fai_path = os.path.join(wk_dir, 'temp1.fa.fai')
    fai_names = {i.split('\t')[0]: '' for i in open(fai_path, 'r').read().splitlines()}
    missed_reads = 0
    for i in index2reads:
        for r in index2reads[i].split(','):
            if r.split('~')[0] in fai_names:
                try:
                    left = str(region_dict[r.strip('\n')][0])
                    right = str(int(region_dict[r.strip('\n')][1]))
                    if r.split('~')[0] + ':' + left + '-' + right not in read2index:
                        bed_str += r.split('~')[0] + '\t' + left + '\t' + right + '\t' + r.strip('\n') + '\n'
                        read2index[r.strip('\n')] = i
                except KeyError:  ## Some reads might not be previously assigned as INS (e.g. TDupl), hence not picked up by region_dict
                    pass
            else:
                missed_reads += 1
    log_missing_reads(missed_reads)
    bed = BedTool(bed_str, from_string=True)
    # Extract fasta for each insertion region
    fasta_path = os.path.join(wk_dir, 'temp1.fa')
    fasta = bed.sequence(fi=fasta_path, name=True)
    _ = fasta.save_seqs(os.path.join(wk_dir, 'ins_seq.fa'))
    return read2index

def log_missing_reads(missed_reads):
    if missed_reads > 0:
        logging.warning("Could not analyze INS sequence of %i reads due to absence of FASTA. Please check if all reads in BAM have a primary alignment." % missed_reads)

def parse_bam_te(bam, wk_dir, read2index, L1_90_len=5500, Alu_90_len=270):
    save = pysam.set_verbosity(0)
    sam = pysam.AlignmentFile(bam, "rb")
    pysam.set_verbosity(save)
    index2te = {}
    unmap_fasta = open(os.path.join(wk_dir, 'ins_seq_te-unmap.fa'), 'w')
    for seg in sam:
        flag = seg.flag
        qname = seg.query_name.split('::')[0]
        if flag == 4:
            unmap_fasta.write('>' + seg.query_name + '\n' + seg.query_sequence + '\n')
        elif flag in (0, 16):
            ref_aligned_len = seg.reference_end - seg.reference_start
            if seg.reference_name.rsplit('_', 1)[1].startswith('L1'):
                if int(ref_aligned_len) >= L1_90_len:  # 90% of L1
                    pass
            elif seg.reference_name.rsplit('_', 1)[1].startswith('Alu'):
                if int(ref_aligned_len) >= Alu_90_len:  # 90% of Alu
                    pass
            else:
                continue
            if read2index[qname] in index2te:
                index2te[read2index[qname]].add(seg.reference_name.rsplit('_', 1)[1])
            else:
                index2te[read2index[qname]] = set()
                index2te[read2index[qname]].add(seg.reference_name.rsplit('_', 1)[1])
    unmap_fasta.close()
    return index2te

def parse_bam_ref(bam, ins_coord_dict, read2index):
    save = pysam.set_verbosity(0)
    sam = pysam.AlignmentFile(bam, "rb")
    pysam.set_verbosity(save)
    index2dup = {}
    dup_coord = {}
    for seg in sam:
        flag = seg.flag
        if flag in (0, 16):
            qname = seg.query_name.split('::')[0]
            dup, coord = overlap_dup(ins_coord_dict[qname], seg)
            if dup:
                dup_coord[qname] = coord
                if read2index[qname] in index2dup:
                    index2dup[read2index[qname]] += 1
                else:
                    index2dup[read2index[qname]] = 1
    return index2dup, dup_coord

def overlap_dup(supposed_coord, seg, buffer=20):
    if seg.reference_name == supposed_coord[0]:
        if supposed_coord[1] - buffer <= seg.reference_start and seg.reference_end <= supposed_coord[2] + buffer:
            return True, seg.reference_name + ':' + str(seg.reference_start) + '-' +  str(seg.reference_end)
        else:
            return False, ''
    else:
        return False, ''

# Add originally TDupl reads
def update_dups(dups, index2dup, index2reads):
    for i in index2reads:
        for r in index2reads[i].split(','):
            if r in dups:
                try:
                    index2dup[i] += 1
                except KeyError:
                    pass
    return index2dup

# Find TDupls based on proportions of dup reads
def filter_dup(index2dup, index2reads, threshold=0.75):
    dup_indexes = {}
    fail_dups = {}
    for i in index2dup:
        if index2dup[i]/len(index2reads[i].split(',')) >= threshold:
            dup_indexes[i] = round(index2dup[i]/len(index2reads[i].split(',')), 3)
        else:
            # fail_dups[i] = round(index2dup[i]/len(index2reads[i].split(',')), 3)
            fail_dups[i] = ''
    return dup_indexes, fail_dups

def update_out_nn(dup_indexes, out_nn, dup_coord):
    new_out_nn = []
    for i in out_nn:
        if i.split('\t')[6].split('~')[0] in dup_indexes:
            index_coord = i.split('\t')[6].split('~')[0] + '~' + get_coord(dup_coord, i)
            new_out_nn.append('\t'.join(i.split('\t')[0:3]) + '\tTDupl 99.99~' + i.split('\t')[3].split('~')[1] + '\t' + '\t'.join(i.split('\t')[4:6]) + '\t' + index_coord + '\t' + '\t'.join(i.split('\t')[7:]))
        else:
            new_out_nn.append(i)
    return new_out_nn

def get_coord(dup_coord, i):
    try:
        return dup_coord[i.split('\t')[8]]
    except KeyError:
        for r in i.split('\t')[11].split(','):
            if r in dup_coord:
                return dup_coord[r]




"""
Functions to calculate the upper limit for long-read SV depth coverage.

Copyright (C) 2021 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""


import os
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
from pybedtools import BedTool
from scipy.interpolate import make_interp_spline, BSpline


# Generate upper overlap limit, depth of coverage, and coverage curve plot
def ovl_upper(total_gsize, contig_len_dict, basecov, beddata, wk_dir):
    # matplotlib.use('Agg')
    bed_str = normalbed(beddata)
    bed = BedTool(bed_str, from_string=True)
    bed = bed.sort()
    n = ngenerate(total_gsize)
    gsize_path = make_gsize(contig_len_dict, wk_dir)
    x = BedTool()
    ran_bed = x.random(l=100, n=n, seed=3, g=gsize_path)
    ran_bed = ran_bed.sort()
    insect = ran_bed.intersect(bed, wa=True, wb=True)
    cov_list = []
    for line in insect:
        line = list(line)
        cov_list.append(line[3])
    cov_dict = {}
    for key in cov_list:
        try:
            cov_dict[key] += 1
        except KeyError:
            cov_dict[key] = 1
    data = []
    for key in cov_dict:
        data.append(cov_dict[key])
    data = [float(i) for i in data]
    zerolist = [0.0]*(n-len(data))
    data2 = data + zerolist
    med = np.median(data2)
    medad = mad(data2)
    curve(data2, n, round((medad*6) + med, 0), wk_dir)
    # Ignore 0 depth regions
    depth = round(np.mean(data), 2)
    # depth = round(float(basecov)/total_gsize, 2)
    med = np.median(data) #
    medad = mad(data) #
    maxovl = max(round((medad * 4) + med, 1), 10)  # minimum overlap threshold is set at 10
    maxovl3 = max(round((medad * 3) + med, 1), 10)
    # me = np.mean(data2)
    # std = np.std(data2)
    return maxovl, depth, maxovl3


# Make genome size file
def make_gsize(contig_len_dict, wk_dir):
    path = os.path.join(wk_dir, 'genome.sizes')
    data = open(path, 'w')
    tmp = []
    for contig in contig_len_dict:
        # checkcontigname(contig)
        tmp.append(contig + '\t' + str(contig_len_dict[contig]))
    data.write('\n'.join(tmp) + '\n')
    data.close()
    return path


# Generate number of genomic points
def ngenerate(gsize):
    if gsize > 10000:
        return 10000
    else:
        return int(gsize*0.9)


# Function to generate bed file from beddata
def normalbed(beddata):
    totalbed = []
    for line in beddata:
        # coord1 = int(line.split('\t')[1])
        # coord2 = int(line.split('\t')[1]) + int(line.split('\t')[2])
        # totalbed.append(line.split('\t')[0] + '\t' + str(coord1) + '\t' + str(coord2) + '\t' + line.split('\t')[4])
        totalbed.append(line[0] + '\t' + str(line[1]) + '\t' + str(line[2]) + '\t' + line[3])
    bed_str = '\n'.join(totalbed)
    return bed_str


# Median absolute deviation
def mad(x):
    b = 1.4826
    return b*np.mean(abs(x-np.median(x)))


# Plot curve
def curve(data, n, upper_limit, wk_dir):
    c = range(max(int(upper_limit), 10))
    p = []
    for i in c:
        p.append(data.count(i))
    p.append(n - sum(p))
    y = np.array([(float(z)/n) for z in p])
    # theoretical = [0.0915, 0.0441, 0.1032, 0.1498, 0.1739, 0.1626,
    # 0.1132, 0.0808, 0.0412, 0.0247, 0.0097, 0.0028, 0.0015,
    # 0.0006, 0.0002, 0.0002, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    c = np.array(range(max(int(upper_limit), 10)+1))
    xnew = np.linspace(c.min(), c.max(), 100)
    spl = make_interp_spline(c, y)
    smooth = spl(xnew)
    params = {'axes.labelsize': 14, 'axes.titlesize': 17, 'legend.fontsize': 10, 'xtick.labelsize': 12, 'ytick.labelsize': 12,
              'font.family': 'sans-serif'}
    matplotlib.rcParams.update(params)
    fig = plt.figure(figsize=(8, 6))
    fig.patch.set_facecolor('#f6f7f9')
    ax = fig.add_subplot(111)
    ax.plot(xnew, smooth, color='#403f7d', linewidth=2.0)
    ax.set_facecolor('#ebebff')
    plt.text(int(upper_limit), y[-1], '>=' + str(int(upper_limit)))
    ax.grid(color='w', linestyle='-', linewidth=1)
    vals = ax.get_yticks().tolist()
    ax.yaxis.set_major_locator(mticker.FixedLocator(vals))
    ax.set_yticklabels(['{:,.1%}'.format(x) for x in vals])
    plt.ylabel('Percentage')
    plt.xlabel('Depth of coverage')
    ax.legend(['Input sequencing data'], loc='upper right', ncol=1, fancybox=True)
    plt.savefig(os.path.join(wk_dir, 'fig', 'depth_of_coverage.png'),
                bbox_inches='tight',
                dpi=100,
                facecolor=fig.get_facecolor(),
                edgecolor='none')



"""
Parses BAM file and analyse CIGARS

Copyright (C) 2021 Tham Cheng Yong

This file is part of NanoVar.

NanoVar is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NanoVar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with NanoVar.  If not, see <https://www.gnu.org/licenses/>.
"""

import pysam
import os
import logging
from .nv_detect_algo import sv_detect
from .nv_parser import entry_parser, breakpoint_parser


def bam_parse(object sam, unsigned int minlen, float splitpct, unsigned int minalign, str wk_dir, filter_file, contig_omit):
    cdef:
        unsigned int readlen, rstart, rend, flag, qlen, nm, seed
        unsigned long long basecov
        float ovlt
        int total_score
        str qname
        str rname
        list qseg, sseg, del_list, ins_list, cigar_tup, total_subdata, total_lines, contig_collect, total_out, sig_index, detect_out, beddata
        set add_sup_seq
        bint adv
        object seg
        #int save = pysam.set_verbosity(0)  # Suppress BAM index missing warning
        #object sam = pysam.AlignmentFile(bam, "rb")
        dict repeat_dict = {}
        dict rlendict = {}
        dict main_dict = {}
        dict parse_dict = {}
        dict gapdict = {}
        dict fail_qual_seq = {}
        dict fail_qual_len = {}
    #pysam.set_verbosity(save)  # Revert verbosity level
    seed = 0
    basecov = 0
    ovlt = 0.9  # Set overlap tolerance
    sig_index = list(range(0, 40, 2))
    fasta = open(os.path.join(wk_dir, 'temp1.fa'), 'w')
    fasta2 = open(os.path.join(wk_dir, 'temp2.fa'), 'w')
    beddata = []
    add_sup_seq = set()
    for seg in sam:
        flag = seg.flag
        qname = seg.query_name
        if flag == 4:
            fasta2.write('>' + qname + '\n' + seg.query_sequence + '\n')
            fasta.write('>' + qname + '\n' + seg.query_sequence + '\n')
            # rlendict[qname] = len(seg.query_sequence)
            continue
        elif flag in (256, 272):  # Skip secondary alignments but save fasta if not done yet
            continue
        if seg.mapping_quality > 0 and seg.mapping_quality < 30:
            if flag in (0, 16):
                fail_qual_seq[qname] = seg.query_sequence
                fail_qual_len[qname] = seg.infer_read_length()
            continue
        rname = seg.reference_name
        readlen = seg.infer_read_length()
        rstart = seg.reference_start
        rend = seg.reference_end
        qlen = seg.query_alignment_length
        nm = seg.get_tag('NM')
        total_score = seg.get_tag('AS')
        cigar_tup = seg.cigartuples
        adv, qseg, sseg, del_list, ins_list = read_cigar(cigar_tup, minlen, splitpct, rstart, rend, readlen)
        beddata.append([rname, rstart, rend, qname])  # 0, 16, 2048, 2064
        if flag in (0, 16):
            try:
                if repeat_dict[qname]:
                    pass
            except KeyError:
                repeat_dict[qname] = ''
                fasta.write('>' + qname + '\n' + seg.query_sequence + '\n')
                rlendict[qname] = readlen
        elif flag in (2048, 2064):
            try:
                if repeat_dict[qname]:
                    pass
            except KeyError:
                add_sup_seq.add(qname)
        try:  # Save data for flags 0, 16, 2048, 2064
            main_dict[qname].append((adv, qname, rname, rstart, rend, readlen, qlen, flag, nm, total_score, qseg, sseg, del_list,
            ins_list))
        except KeyError:
            main_dict[qname] = []
            main_dict[qname].append((adv, qname, rname, rstart, rend, readlen, qlen, flag, nm, total_score, qseg, sseg, del_list,
            ins_list))
    for sup in add_sup_seq:
        try:
            if repeat_dict[sup]:
                pass
        except KeyError:
            # Some supplementary alignments do not have primary alignments (in the case of a subset BAM), thus sequence not found and read sequence excluded
            if sup in fail_qual_seq:
                repeat_dict[sup] = ''
                fasta.write('>' + sup + '\n' + fail_qual_seq[sup] + '\n')
                rlendict[sup] = fail_qual_len[sup]
    fasta.close()
    fasta2.close()
    total_subdata, total_lines, contig_collect, total_out, detect_out = [], [], [], [], []
    # Make gap dictionary
    gapdict = makegapdict(filter_file, contig_omit)
    for qname in main_dict:
        if len(main_dict[qname]) == 1:  # Single alignment read
            if not main_dict[qname][0][0]:  # if no sub-segments and not clipped read
                aln = main_dict[qname][0]
                qname = aln[1]
                rname = aln[2]
                rstart = aln[3]
                rend = aln[4]
                total_subdata.append(info_parse_simple(qname, rname, rstart, rend, minalign))
                # Add to base coverage
                basecov += rend - rstart
            else:  # if multiple sub-segments or clipped read
                seed += 1
                aln = main_dict[qname][0]
                qname = aln[1]
                rname = aln[2]
                readlen = aln[5]
                qlen = aln[6]
                flag = aln[7]
                nm = aln[8]
                total_score = aln[9]
                qseg = aln[10]
                sseg = aln[11]
                del_list = aln[12]
                ins_list = aln[13]     
                total_lines, contig_collect = info_parse(qname, rname, readlen, qlen, flag, nm, total_score, qseg, sseg,
                del_list, ins_list, minalign)
                contig_collect = sorted(set(contig_collect))
                lines_sort = sorted(sorted(total_lines, key=lambda x: x[1], reverse=True), key=lambda y: y[0])
                temp1 = [tup[2] for tup in lines_sort]
                # Parse entries and correct overlap alignments
                subdata = entry_parser(temp1, contig_collect, ovlt)
                for entry in subdata:
                    # total_subdata.append('\t'.join(entry.split('\t')[0:5]))
                    total_subdata.append(entry)
                    # Add to base coverage
                    basecov += int(entry.split('\t')[2])
                # SV detection
                out1, out2 = sv_detect(subdata, splitpct, minalign, gapdict)
                if out2 == '':
                    pass
                else:
                    # Save for debug
                    detect_out.append(out2)
                    # Parse breakpoints
                    final = breakpoint_parser(out2, minlen, sig_index, seed, 'mm')
                    total_out.extend(final)
                    for i in final:
                        parse_dict[i.split('\t')[8]] = '\t'.join(i.split('\t')[0:5]) + '\tmm\t' + '\t'.join(i.split('\t')[6:])
        else:  # Multiple alignment read
            seed += 1
            total_lines = []
            contig_collect = []
            for aln in main_dict[qname]:
                qname = aln[1]
                rname = aln[2]
                readlen = aln[5]
                qlen = aln[6]
                flag = aln[7]
                nm = aln[8]
                total_score = aln[9]
                qseg = aln[10]
                sseg = aln[11]
                del_list = aln[12]
                ins_list = aln[13]
                lines, contigs = info_parse(qname, rname, readlen, qlen, flag, nm, total_score, qseg, sseg,
                del_list, ins_list, minalign)
                total_lines.extend(lines)
                contig_collect.extend(contigs)
            contig_collect = sorted(set(contig_collect))
            lines_sort = sorted(sorted(total_lines, key=lambda x: x[1], reverse=True), key=lambda y: y[0])
            temp1 = [tup[2] for tup in lines_sort]
            # Parse entries and correct overlap alignments
            subdata = entry_parser(temp1, contig_collect, ovlt)
            for entry in subdata:
                # total_subdata.append('\t'.join(entry.split('\t')[0:5]))
                total_subdata.append(entry)
                # Add to base coverage
                basecov += int(entry.split('\t')[2])
            # SV detection
            out1, out2 = sv_detect(subdata, splitpct, minalign, gapdict)
            if out2 == '':
                pass
            else:
                # Save for debug
                detect_out.append(out2)
                # Parse breakpoints
                final = breakpoint_parser(out2, minlen, sig_index, seed, 'mm')
                total_out.extend(final)
                for i in final:
                    parse_dict[i.split('\t')[8]] = '\t'.join(i.split('\t')[0:5]) + '\tmm\t' + '\t'.join(i.split('\t')[6:])
    return total_subdata, total_out, basecov, parse_dict, rlendict, len(repeat_dict), detect_out, seed, beddata


# Analyze CIGAR for Indels in segment and return read advancement call
cdef read_cigar(list cigar_tup, float minlen, float splitpct, unsigned int rstart, unsigned int rend, unsigned int readlen):
    cdef unsigned int qstart = 1
    cdef unsigned int softstart, softend, qlast, sstart, slast, dels, ins, b, v, t, move
    cdef list qseg = []
    cdef list sseg = []
    cdef list del_list = []
    cdef list ins_list = []

    if cigar_tup[0][0] == 4 or cigar_tup[0][0] == 5:
        qstart += cigar_tup[0][1]
        softstart = cigar_tup[0][1]
    else:
        softstart = 0
    if cigar_tup[-1][0] == 4 or cigar_tup[-1][0] == 5:
        softend = cigar_tup[-1][1]
    else:
        softend = 0
    qlast, dels, ins = 0, 0, 0
    sstart = rstart
    slast = sstart - 1
    for t, move in cigar_tup:
        if t in (4, 5):  # S or H
            qlast += move
        elif t in (0, 7, 8):  # M or = or X
            qlast += move
            slast += move
        elif t == 2:  # D
            dels += move
            if move > minlen:
                qseg.append([qstart, qlast])
                qstart = qlast + 1
                sseg.append([sstart, slast])
                slast += move
                sstart = slast + 1
                del_list.append(dels)
                ins_list.append(ins)
                dels = 0
                ins = 0
            else:
                slast += move
        elif t == 1:  # I
            ins += move
            if move > minlen:
                qseg.append([qstart, qlast])
                qlast += move
                qstart = qlast + 1
                sseg.append([sstart, slast])
                sstart = slast + 1
                del_list.append(dels)
                ins_list.append(ins)
                dels = 0
                ins = 0
            else:
                qlast += move
        else:
            logging.critical('Error: Unrecognized CIGAR translated symbol "%s"' % t)
            raise Exception('Error: Unrecognized CIGAR translated symbol "%s"' % t)
    # Last segment
    qseg.append([qstart, qlast - softend])
    sseg.append([sstart, slast])
    del_list.append(dels)
    ins_list.append(ins)
    if len(qseg) == 1:
        if softstart > splitpct * readlen or softend > splitpct * readlen:
            return True, qseg, sseg, del_list, ins_list
        else:
            return False, qseg, sseg, del_list, ins_list
    else:
        return True, qseg, sseg, del_list, ins_list



def info_parse(qname, rname, readlen, qlen, flag, nm, total_score, qseg, sseg, del_list, ins_list, minlen=200):
    lines = []
    total_del = sum([x for x in del_list])
    total_ins = sum([x for x in ins_list])
    total_gap = total_del + total_ins
    total_mismatch = nm - total_gap
    nsegs = len(qseg)
    contig_collect = []
    for i in range(nsegs):
        substart = sseg[i][0]
        substretch = sseg[i][1] - sseg[i][0]
        if substretch < 20:
            continue
        qstart, qstretch, strand = query_sign(qseg[i][0], qseg[i][1], flag, readlen)
        ndel = del_list[i]
        nins = ins_list[i]
        gaps = ndel + nins
        subratio = round(qstretch / qlen, 3)
        mismatch = int(round(total_mismatch * subratio, 0))
        matches = max(qstretch - mismatch - nins, 1)
        pident = min(round(matches * 100 / substretch, 2), 100.00)
        score = int(round(total_score * subratio, 0))
        priority = align_priority(flag)
        line = "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s" % (rname, substart, substretch, '+',
                                                                           qname, qstart, qstretch, strand,
                                                                           readlen, 0.0, score, pident, mismatch, gaps)
        contig_collect.append(rname)
        lines.append((priority, score, line))
    return lines, contig_collect

# Parse segments without breakends
def info_parse_simple(qname, rname, substart, rend, minlen=200):
    substretch = rend - substart
    if substretch < minlen:
        pass
    line = "%s\t%s\t%s\t%s\t%s" % (rname, substart, substretch, '+', qname)
    return line

# Parse query start and end values according to read direction
def query_sign(start, end, flag, readlen):
    if flag in (0, 2048, 256):
        return start, end - start, '+'
    elif flag in (16, 2064, 272):
        return readlen - end, end - start, '-'
    else:
        return start, end - start, '+'

# Set alignment priority according to flag
def align_priority(flag):
    if flag in (0, 16):
        return 1
    elif flag in (2048, 2064):
        return 2
    else:
        return 3


# Create genome gap dictionary
def makegapdict(gap_path, contig_omit):
    gapdict = {}
    if gap_path is not None:
        rgapdata = open(gap_path, 'r').read().splitlines()
        for line in rgapdata:
            try:
                gapdict[line.split('\t')[0]].append([int(line.split('\t')[1]), int(line.split('\t')[2])])
            except KeyError:
                gapdict[line.split('\t')[0]] = [[int(line.split('\t')[1]), int(line.split('\t')[2])]]
    for contig in contig_omit:
        try:
            gapdict[contig].append(contig_omit[contig])
        except KeyError:
            gapdict[contig] = [contig_omit[contig]]
    if gapdict:  # Gap file is used or reference genome contigs contain invalid symbols
        logging.info('Gap dictionary successfully loaded')
        return gapdict
    else:
        logging.info('Gap dictionary not loaded.')
        return None
